# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gio, GLib

from gettext import gettext as _
from urllib.parse import urlparse

from eolie.menu_languages import LanguagesMenu
from eolie.menu_scripts import ScriptsMenu
from eolie.define import App
from eolie.utils import get_safe_netloc


class ToolbarMenu(Gtk.PopoverMenu):
    """
        Gtk.PopoverMenu showing main menu
    """

    def __init__(self, uri, window):
        """
            Init self
            @param uri as str
            @param window as Window
        """
        Gtk.PopoverMenu.__init__(self)
        self.__uri = uri
        self.__window = window
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/ToolbarMenu.ui")
        fullscreen_button = builder.get_object("fullscreen_button")
        if self.__window.is_fullscreen:
            fullscreen_button.set_active(True)
            fullscreen_button.set_tooltip_text(_("Leave fullscreen"))
        else:
            fullscreen_button.set_tooltip_text(_("Enter fullscreen"))
        builder.connect_signals(self)
        widget = builder.get_object("widget")
        self.add(widget)
        webview = self.__window.container.webview
        builder.get_object("default_zoom_button").set_label(
            "{} %".format(int(webview.get_zoom_level() * 100)))
        netloc = get_safe_netloc(uri)
        parsed = urlparse(uri)
        builder.get_object("domain_label").set_text(netloc)
        if parsed.scheme in ["http", "https"]:
            # Add blocker actions
            for blocker in ["block-ads", "block-popups",
                            "block-images", "block-medias"]:
                if not App().settings.get_value(blocker):
                    continue
                builder.get_object(blocker).show()
                content_blocker = App().get_content_blocker(blocker)
                exception = content_blocker.exceptions.is_domain_exception(
                    netloc)
                action = Gio.SimpleAction.new_stateful(
                    "%s-exception" % blocker,
                    None,
                    GLib.Variant.new_boolean(exception))
                action.connect("change-state",
                               self.__on_exception_change_state,
                               netloc,
                               blocker)
                window.add_action(action)
            # Audio policy
            netloc_audio = App().websettings.get("audio", uri)
            builder.get_object("audio_policy").show()
            action = Gio.SimpleAction.new_stateful(
                    "audio-policy",
                    None,
                    GLib.Variant.new_boolean(netloc_audio))
            action.connect("change-state",
                           self.__on_audio_policy_change_state,
                           uri)
            window.add_action(action)
            # Night mode
            night_mode = App().settings.get_value("night-mode")
            netloc_night_mode = App().websettings.get("night_mode", uri)
            builder.get_object("night_mode").show()
            enabled = night_mode and netloc_night_mode in [1, None]
            action = Gio.SimpleAction.new_stateful(
                    "night-mode",
                    None,
                    GLib.Variant.new_boolean(enabled))
            action.connect("change-state",
                           self.__on_night_mode_change_state,
                           uri)
            window.add_action(action)
            # Scripts
            if App().settings.get_value("block-scripts"):
                builder.get_object("scripts").show()
                scripts_menu = ScriptsMenu(uri, self.__window)
                scripts_menu.show()
                self.add(scripts_menu)
                self.child_set_property(scripts_menu, "submenu", "scripts")
            # Languages
            builder.get_object("spell-checking").show()
            languages_menu = LanguagesMenu(uri)
            languages_menu.show()
            self.add(languages_menu)
            self.child_set_property(languages_menu, "submenu", "languages")

#######################
# PROTECTED           #
#######################
    def _on_print_button_clicked(self, button):
        """
            Print current page
            @param button as Gtk.button
        """
        button.get_ancestor(Gtk.Popover).hide()
        action = self.__window.lookup_action("shortcut")
        action.activate(GLib.Variant("s", "print"))

    def _on_save_button_clicked(self, button):
        """
            Save current page
            @param button as Gtk.Button
        """
        button.get_ancestor(Gtk.Popover).hide()
        action = self.__window.lookup_action("shortcut")
        action.activate(GLib.Variant("s", "save"))

    def _on_zoom_button_clicked(self, button):
        """
            Zoom current page
            @param button as Gtk.Button
        """
        webview = self.__window.container.webview
        current = webview.zoom_in()
        button.set_label("{} %".format(current))

    def _on_unzoom_button_clicked(self, button):
        """
            Unzoom current page
            @param button as Gtk.Button
        """
        webview = self.__window.container.webview
        current = webview.zoom_out()
        button.set_label("{} %".format(current))

    def _on_fullscreen_button_toggled(self, button):
        """
            Restore default zoom level
            @param button as Gtk.ToggleButton
        """
        button.get_ancestor(Gtk.Popover).hide()
        if button.get_active():
            if not self.__window.is_fullscreen:
                self.__window.fullscreen()
        else:
            if self.__window.is_fullscreen:
                self.__window.unfullscreen()

    def _on_default_zoom_button_clicked(self, button):
        """
            Restore default zoom level
            @param button as Gtk.Button
        """
        webview = self.__window.container.webview
        current = webview.zoom_default()
        button.set_label("{} %".format(current))

#######################
# PRIVATE             #
#######################
    def __on_exception_change_state(self, action, param, domain, blocker):
        """
            Set option value
            @param action as Gio.SimpleAction
            @param param as GLib.Variant
            @param domain as str
            @param blocker as str
        """
        action.set_state(param)
        content_blocker = App().get_content_blocker(blocker)
        if content_blocker.exceptions.is_domain_exception(domain):
            content_blocker.exceptions.remove_domain_exception(domain)
        else:
            content_blocker.exceptions.add_domain_exception(domain)
        content_blocker.exceptions.save()
        content_blocker.update()
        self.__window.container.webview.reload()

    def __on_night_mode_change_state(self, action, param, uri):
        """
            Set night mode value
            @param action as Gio.SimpleAction
            @param param as GLib.Variant
            @param uri as str
        """
        action.set_state(param)
        App().websettings.set("night_mode", uri, param.get_boolean())
        self.__window.container.webview.night_mode()

    def __on_audio_policy_change_state(self, action, param, uri):
        """
            Set audio policy for URI
            @param action as Gio.SimpleAction
            @param param as GLib.Variant
            @param uri as str
        """
        action.set_state(param)
        value = param.get_boolean()
        App().websettings.set("audio", uri, value)
        self.__window.container.webview.update_sound_policy()



# Copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from eolie.define import App


class BookmarkRatingWidget(Gtk.Bin):
    """
        Rate widget
    """

    def __init__(self, bookmark_id):
        """
            Init widget
            @param bookmark_id as int
        """
        Gtk.Bin.__init__(self)
        self.__bookmark_id = bookmark_id
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/BookmarkRatingWidget.ui")
        builder.connect_signals(self)
        self.set_property("valign", Gtk.Align.CENTER)
        self.set_property("halign", Gtk.Align.END)
        self._hearts = []
        self._hearts.append(builder.get_object("heart0"))
        self._hearts.append(builder.get_object("heart1"))
        self._hearts.append(builder.get_object("heart2"))
        self._hearts.append(builder.get_object("heart3"))
        self._hearts.append(builder.get_object("heart4"))
        self._on_leave_notify(None, None)
        self.add(builder.get_object("widget"))

#######################
# PROTECTED           #
#######################
    def _on_enter_notify(self, widget, event):
        """
            On enter notify, change heart opacity
            @param widget as Gtk.EventBox
            @param event as Gdk.Event
        """
        event_heart = widget.get_children()[0]
        # First heart is hidden (used to clear score)
        if event_heart.get_opacity() == 0.0:
            found = True
        else:
            found = False
        for heart in self._hearts:
            if found:
                heart.set_opacity(0.2)
            else:
                heart.set_opacity(0.8)
            if heart == event_heart:
                found = True

    def _on_leave_notify(self, widget, event):
        """
            On leave notify, change heart opacity
            @param widget as Gtk.EventBox (can be None)
            @param event as Gdk.Event (can be None)
        """
        rate = self.__get_rate()
        if rate < 1:
            for i in range(5):
                self._hearts[i].set_opacity(0.2)
        else:
            heart = self.__heart_from_rate(rate)
            # Select wanted heart
            for idx in range(0, heart):
                widget = self._hearts[idx]
                widget.set_opacity(0.8)
            # Unselect others
            for idx in range(heart, 5):
                self._hearts[idx].set_opacity(0.2)

    def _on_button_press(self, widget, event):
        """
            On button press, set album popularity
            @param widget as Gtk.EventBox
            @param event as Gdk.Event
        """
        rate = self.__get_rate()
        max_heart = self.__heart_from_rate(rate)
        event_heart = widget.get_children()[0]
        if event_heart in self._hearts:
            position = self._hearts.index(event_heart)
        else:
            position = -1
        new_rate = position + 1
        if new_rate == 0 or new_rate == max_heart:
            App().bookmarks.set_popularity(self.__bookmark_id, 0)
            self._on_leave_notify(None, None)
        else:
            avg_popularity = App().bookmarks.get_avg_popularity()
            best_popularity = App().bookmarks.get_higher_popularity()
            popularity = int((new_rate * avg_popularity / 5) + 0.5)
            # Special case, if new_rate == 5, try to move bookmark near
            # most popular bookmark
            if new_rate == 5:
                popularity = (popularity + best_popularity) / 2
            App().bookmarks.set_popularity(self.__bookmark_id, popularity)
        return True

#######################
# PRIVATE             #
#######################
    def __get_rate(self):
        """
            Get bookmark rate: popularity related to avarage popularity
        """
        avg_popularity = App().bookmarks.get_avg_popularity()
        popularity = App().bookmarks.get_popularity(self.__bookmark_id)
        return popularity * 5 / avg_popularity + 0.5

    def __heart_from_rate(self, rate):
        """
            Calculate hearts from rate
            @param rate as double
            @return int
        """
        heart = min(5, int(rate))
        return heart



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib


class SmoothProgressBar(Gtk.ProgressBar):
    """
        Gtk.ProgressBar with full progression
        If faction is 0.1 and you set it to 0.5, it will be udpated
        to show all steps
    """

    def __init__(self):
        """
            Init widget
        """
        Gtk.ProgressBar.__init__(self)
        self.__timeout_id = None
        self.set_property("valign", Gtk.Align.END)
        self.get_style_context().add_class("progressbar")

    def set_fraction(self, fraction):
        """
            Set fraction smoothly
            @param fraction as float
        """
        if self.__timeout_id is not None:
            GLib.source_remove(self.__timeout_id)
        self.__set_fraction(fraction)

    def hide(self):
        """
            Hide widget and reset fraction
        """
        Gtk.ProgressBar.hide(self)
        Gtk.ProgressBar.set_fraction(self, 0)
        if self.__timeout_id is not None:
            GLib.source_remove(self.__timeout_id)
            self.__timeout_id = None

#######################
# PRIVATE             #
#######################
    def __set_fraction(self, fraction):
        """
            Set fraction smoothly
            @param fraction as float
        """
        Gtk.ProgressBar.show(self)
        self.__timeout_id = None
        current = self.get_fraction()
        if fraction < current:
            progress = fraction
        else:
            if current > 0.95:
                delta = 1.0 - current
            elif fraction - current > 0.5 or fraction == 1.0:
                delta = (fraction - current) / 10
            else:
                delta = (fraction - current) / 100
            progress = current + delta
        Gtk.ProgressBar.set_fraction(self, progress)
        if progress < 1.0:
            self.__timeout_id = GLib.timeout_add(10,
                                                 self.__set_fraction,
                                                 fraction)
        else:
            self.__timeout_id = GLib.timeout_add(500, self.hide)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib

from hashlib import sha256

from eolie.define import App


class HistoryMenu(Gio.Menu):
    """
        Menu showing closed page
    """

    def __init__(self, items, window):
        """
            Init menu
            @param items as [WebKit2.BackForwardListItem]
        """
        Gio.Menu.__init__(self)
        self.__window = window
        self.__actions = []
        for item in items[:10]:
            uri = item.get_uri()
            if uri is None:
                continue
            title = item.get_title()
            if not title:
                title = uri
            encoded = "HISTORY_" + sha256(uri.encode("utf-8")).hexdigest()
            action = Gio.SimpleAction(name=encoded)
            window.add_action(action)
            self.__actions.append(encoded)
            action.connect('activate',
                           self.__on_action_activate,
                           item)
            item = Gio.MenuItem.new(title, "win.%s" % encoded)
            item.set_attribute_value("uri", GLib.Variant("s", uri))
            if uri == "populars://":
                item.set_icon(Gio.ThemedIcon.new("emote-love-symbolic"))
            else:
                # Try to set icon
                favicon_path = App().art.get_favicon_path(uri)
                if favicon_path is not None:
                    f = Gio.File.new_for_path(favicon_path)
                    icon = Gio.FileIcon.new(f)
                    if icon is not None:
                        item.set_icon(icon)
                else:
                    item.set_icon(Gio.ThemedIcon.new("web-browser-symbolic"))
            self.append_item(item)

    def clean(self):
        """
            Clean menu
        """
        for action in self.__actions:
            self.__window.remove_action(action)

#######################
# PRIVATE             #
#######################
    def __on_action_activate(self, action, variant, item):
        """
            Load history
            @param Gio.SimpleAction
            @param GVariant
            @param item as WebKit2.BackForwardListItem
        """
        App().active_window.\
            container.webview.go_to_back_forward_list_item(item)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib, Gtk

from gettext import gettext as _

from eolie.define import App, LoadingType


class ApplicationMenu(Gio.Menu):
    """
        Show application menu
    """

    def __init__(self, window):
        """
            Init menu
            @param window as Window
        """
        Gio.Menu.__init__(self)
        self.__window = window
        # Main actions menu
        action = Gio.SimpleAction(name="new-private")
        App().add_action(action)
        action.connect("activate",
                       self.__on_private_clicked)
        self.append(_("New private page"), "app.new-private")

        # Blockers submenu
        blockers_menu = Gio.Menu()
        for (label, blocker) in [
                (_("Block ads"), "block-ads"),
                (_("Block popups"), "block-popups"),
                (_("Block images"), "block-images"),
                (_("Block multimedia"), "block-medias"),
                (_("Block scripts"), "block-scripts")]:
            value = App().settings.get_value(blocker)
            action = Gio.SimpleAction.new_stateful(
                blocker,
                None,
                GLib.Variant.new_boolean(value))
            App().add_action(action)
            action.connect("change-state",
                           self.__on_blocker_change_state,
                           blocker)
            blockers_menu.append(label, "app.%s" % blocker)

        self.append_submenu(_("Blockers"), blockers_menu)

        # Default application menu
        section = Gio.Menu()
        settings_action = Gio.SimpleAction.new("settings", None)
        settings_action.connect("activate", self.__on_settings_activate)
        App().add_action(settings_action)
        section.append(_("_Preferences"), "app.settings")

        show_sidebar = App().settings.get_value("show-sidebar")
        sidebar_action = Gio.SimpleAction.new_stateful(
            "sidebar",
            None,
            GLib.Variant.new_boolean(show_sidebar))
        sidebar_action.connect("change-state", self.__on_sidebar_change_state)
        App().add_action(sidebar_action)
        section.append(_("_Sidebar"), "app.sidebar")

        value = App().settings.get_value("night-mode")
        action = Gio.SimpleAction.new_stateful(
            "night-mode",
            None,
            GLib.Variant.new_boolean(value))
        App().add_action(action)
        action.connect("change-state",
                       self.__on_night_mode_change_state)
        section.append(_("Night mode"), "app.night-mode")

        shortcuts_action = Gio.SimpleAction.new("shortcuts", None)
        shortcuts_action.connect("activate", self.__on_shortcuts_activate)
        App().add_action(shortcuts_action)
        section.append(_("_Keyboard Shortcuts"), "app.shortcuts")
        self.append_section(None, section)

        about_action = Gio.SimpleAction.new("about", None)
        about_action.connect("activate", self.__on_about_activate)
        App().add_action(about_action)
        section.append(_("_About"), "app.about")

#######################
# PRIVATE             #
#######################
    def __on_settings_activate(self, action, param):
        """
            Show settings dialog
            @param action as Gio.SimpleAction
            @param param as GLib.Variant
        """
        from eolie.dialog_settings import SettingsDialog
        dialog = SettingsDialog(self.__window)
        dialog.show()

    def __on_about_activate(self, action, param):
        """
            Setup about dialog
            @param action as Gio.SimpleAction
            @param param as GLib.Variant
        """
        builder = Gtk.Builder()
        builder.add_from_resource('/org/gnome/Eolie/AboutDialog.ui')
        about = builder.get_object('about_dialog')
        about.set_transient_for(self.__window)
        about.connect("response", self.__on_about_activate_response)
        about.show()

    def __on_shortcuts_activate(self, action, param):
        """
            Show shortcuts
            @param action as Gio.SimpleAction
            @param param as GLib.Variant
        """
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/Shortcuts.ui")
        shortcuts = builder.get_object("shortcuts")
        shortcuts.set_transient_for(self.__window)
        shortcuts.show()

    def __on_about_activate_response(self, dialog, response_id):
        """
            Destroy about dialog when closed
            @param dialog as Gtk.Dialog
            @param response ID as int
        """
        dialog.destroy()

    def __on_sidebar_change_state(self, action, value):
        """
            Show/hide sidebar
            @param action as Gio.SimpleAction
            @param value as bool
        """
        action.set_state(value)
        App().settings.set_value("show-sidebar", GLib.Variant("b", value))

    def __on_blocker_change_state(self, action, value, blocker):
        """
            Update blocker value
            @param action as Gio.SimpleAction
            @param value as bool
            @param blocker as str
        """
        action.set_state(value)
        App().settings.set_value(blocker, GLib.Variant("b", value))
        self.__window.container.webview.reload()

    def __on_night_mode_change_state(self, action, value):
        """
            Update night mode value
            @param action as Gio.SimpleAction
            @param value as bool
        """
        action.set_state(value)
        App().settings.set_value("night-mode", GLib.Variant("b", value))
        settings = Gtk.Settings.get_default()
        settings.set_property("gtk-application-prefer-dark-theme", value)
        for window in App().windows:
            GLib.idle_add(window.toolbar.title.entry.update_style)
            for webview in window.container.webviews:
                webview.night_mode()

    def __on_private_clicked(self, action, variant):
        """
            Add a new private view
            @param Gio.SimpleAction
            @param GVariant
        """
        App().active_window.container.add_webview_for_uri(
                                                  App().start_page,
                                                  LoadingType.FOREGROUND,
                                                  True)



# Copyright (c) 2014-2018 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from eolie.sqlcursor import SqlCursor
from eolie.define import Type


class DatabaseUpgrade:
    """
        Manage database schema upgrades
    """

    def __init__(self, t):
        """
            Init object
            @param t as Type
        """
        # Here are schema upgrade, key is database version,
        # value is sql request
        if t == Type.BOOKMARK:
            self.__UPGRADES = {
                1: self.__upgrade_bookmarks_1,
                2: "ALTER TABLE bookmarks ADD startup INT NOT NULL DEFAULT 0",
            }
        elif t == Type.HISTORY:
            self.__UPGRADES = {
                1: "ALTER TABLE history ADD opened INT NOT NULL DEFAULT 0",
                2: "ALTER TABLE history ADD netloc TEXT NOT NULL DEFAULT ''",
                3: "DELETE FROM history WHERE popularity=0",
                4: "DELETE FROM history_atime WHERE NOT EXISTS (SELECT * FROM\
                    history WHERE history.rowid=history_atime.history_id)",
                5: "CREATE INDEX idx_orderby ON history(mtime, popularity)",
                6: "CREATE INDEX idx_where ON history(uri, title)"
            }
        elif t == Type.SETTINGS:
            self.__UPGRADES = {
                1: "ALTER TABLE settings ADD audio INT NOT NULL DEFAULT 1",
            }

    def upgrade(self, db):
        """
            Upgrade db
            @param db as Database
        """
        version = 0
        with SqlCursor(db) as sql:
            result = sql.execute("PRAGMA user_version")
            v = result.fetchone()
            if v is not None:
                version = v[0]
            if version < self.version:
                for i in range(version + 1, self.version + 1):
                    try:
                        if isinstance(self.__UPGRADES[i], str):
                            sql.execute(self.__UPGRADES[i])
                        else:
                            self.__UPGRADES[i](db)
                    except Exception as e:
                        print("DatabaseUpgrade %s failed: %s" % (i, e))
                sql.execute("PRAGMA user_version=%s" % self.version)

    @property
    def version(self):
        """
            Current wanted version
        """
        return len(self.__UPGRADES)

#######################
# PRIVATE             #
#######################
    def __upgrade_bookmarks_1(self, db):
        """
            Remove del column
            @param db as BookmarksDatabase
        """
        create_bookmarks = '''CREATE TABLE bookmarks (
                                           id INTEGER PRIMARY KEY,
                                           title TEXT NOT NULL,
                                           uri TEXT NOT NULL,
                                           popularity INT NOT NULL,
                                           atime REAL NOT NULL,
                                           guid TEXT NOT NULL,
                                           mtime REAL NOT NULL,
                                           position INT DEFAULT 0
                                           )'''
        with SqlCursor(db) as sql:
            sql.execute("ALTER TABLE bookmarks RENAME TO _bookmarks")
            sql.execute(create_bookmarks)
            sql.execute("""INSERT INTO bookmarks (id, title, uri,
                            popularity, atime, guid, mtime, position)
                           SELECT id, title, uri, popularity, atime, guid,
                            mtime, position FROM _bookmarks""")
            sql.execute("DROP TABLE _bookmarks")



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk, GdkPixbuf, Gio, GLib

from hashlib import md5
from time import time
from urllib.parse import urlparse

from eolie.define import EOLIE_CACHE_PATH, ArtSize
from eolie.utils import get_round_surface
from eolie.logger import Logger


class Art:
    """
        Base art manager
    """

    __CACHE_DELTA = 43200

    def __init__(self):
        """
            Init base art
        """
        self.__use_cache = True
        self.__create_cache()

    def disable_cache(self):
        """
            Disable cache
        """
        self.__use_cache = False

    def save_artwork(self, uri, surface, suffix):
        """
            Save artwork for uri with suffix
            @param uri as str
            @param surface as cairo.surface
            @param suffix as str
        """
        try:
            parsed = urlparse(uri)
            if parsed.scheme in ["http", "https"]:
                filepath = self.get_path(uri, suffix)
                pixbuf = Gdk.pixbuf_get_from_surface(surface, 0, 0,
                                                     surface.get_width(),
                                                     surface.get_height())
                pixbuf.savev(filepath, "png", [None], [None])
        except Exception as e:
            Logger.error("Art::save_artwork(): %s", e)

    def get_artwork(self, uri, suffix, scale_factor, width, heigth):
        """
            @param uri as str
            @param suffix as str
            @param scale factor as int
            @param width as int
            @param height as int
            @return cairo.surface
        """
        if not uri:
            return None
        filepath = self.get_path(uri, suffix)
        try:
            if GLib.file_test(filepath, GLib.FileTest.IS_REGULAR):
                pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_scale(filepath,
                                                                 width,
                                                                 heigth,
                                                                 True)
                surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf,
                                                               scale_factor,
                                                               None)
                return surface
        except:
            pass
        return None

    def get_favicon(self, uri, scale_factor):
        """
            @param uri as str
            @param suffix as str
            @param scale factor as int
            @return cairo.surface
        """
        try:
            if not uri:
                return None
            filepath = self.get_favicon_path(uri)
            if filepath is not None and\
                    GLib.file_test(filepath, GLib.FileTest.IS_REGULAR):
                pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_scale(
                    filepath,
                    ArtSize.FAVICON * scale_factor,
                    ArtSize.FAVICON * scale_factor,
                    True)
                surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf,
                                                               scale_factor,
                                                               None)
                surface = get_round_surface(surface,
                                            scale_factor,
                                            ArtSize.FAVICON / 4)
                return surface
        except Exception as e:
            Logger.debug("Art::get_favicon(): %s", e)
        return None

    def get_icon_theme_artwork(self, uri, ephemeral):
        """
            Get artwork from icon theme
            @param uri as str
            @param ephemeral as bool
            @return artwork as str/None
        """
        if ephemeral:
            return "user-not-tracked-symbolic"
        elif uri == "populars:":
            return "emote-love-symbolic"
        elif uri == "about:":
            return "web-browser-symbolic"
        else:
            return None

    def get_favicon_path(self, uri):
        """
            Return favicon cache path for uri
            @param uri as str/None
            @return str/None
        """
        if uri is None:
            return None
        favicon_path = self.get_path(uri, "favicon")
        if favicon_path is not None and\
                GLib.file_test(favicon_path, GLib.FileTest.IS_REGULAR):
            return favicon_path
        return None

    def get_path(self, uri, suffix):
        """
            Return cache image path
            @param uri as str
            @param suffix as str
            @return str/None
        """
        parsed = urlparse(uri)
        if uri is None or not parsed.netloc:
            return None
        path = "%{}{}%".format(parsed.netloc, parsed.path)
        encoded = md5(path.encode("utf-8")).hexdigest()
        filepath = "%s/art/%s_%s.png" % (EOLIE_CACHE_PATH, encoded, suffix)
        return filepath

    def uncache(self, uri, suffix):
        """
            Remove from cache
            @param uri as str
            @param suffix as str
        """
        try:
            f = Gio.File.new_for_path(self.get_path(uri, suffix))
            f.delete()
        except Exception as e:
            Logger.debug("Art::uncache(): %s", e)

    def exists(self, uri, suffix):
        """
            Check if file exists and is cached
            @param uri as str (raise exception if None)
            @param suffix as str
            @return exists as bool
        """
        filepath = self.get_path(uri, suffix)
        if filepath is None:
            return True  # Because we know Lollypop will do nothing on True
        f = Gio.File.new_for_path(filepath)
        exists = f.query_exists()
        if exists and self.__use_cache:
            info = f.query_info('time::modified',
                                Gio.FileQueryInfoFlags.NONE,
                                None)
            mtime = int(info.get_attribute_as_string('time::modified'))
            return time() - mtime < self.__CACHE_DELTA
        else:
            return False

    def vacuum(self):
        """
            Remove artwork older than 1 month
        """
        current_time = time()
        try:
            d = Gio.File.new_for_path("%s/art" % EOLIE_CACHE_PATH)
            children = d.enumerate_children("standard::name",
                                            Gio.FileQueryInfoFlags.NONE,
                                            None)
            for child in children:
                f = children.get_child(child)
                if child.get_file_type() == Gio.FileType.REGULAR:
                    info = f.query_info("time::modified",
                                        Gio.FileQueryInfoFlags.NONE,
                                        None)
                    mtime = info.get_attribute_uint64("time::modified")
                    if current_time - mtime > 31536000:
                        f.delete()
        except Exception as e:
            Logger.error("Art::vacuum(): %s", e)

#######################
# PRIVATE             #
#######################
    def __create_cache(self):
        """
            Create cache dir
        """
        if not GLib.file_test("%s/art" % EOLIE_CACHE_PATH,
                              GLib.FileTest.IS_DIR):
            try:
                GLib.mkdir_with_parents(EOLIE_CACHE_PATH, 0o0750)
                GLib.mkdir_with_parents("%s/art" % EOLIE_CACHE_PATH, 0o0750)
                GLib.mkdir_with_parents("%s/css" % EOLIE_CACHE_PATH, 0o0750)
            except Exception as e:
                Logger.error("Art::__create_cache(): %s", e)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, WebKit2

from gettext import gettext as _


class JavaScriptPopover(Gtk.Popover):
    """
        Show JavaScript message
        @warning: will block current execution
    """

    def __init__(self, dialog, window):
        """
            Init popover
            @param dialog as WebKit2.ScriptDialog
            @param window as window
        """
        Gtk.Popover.__init__(self)
        self.set_modal(False)
        window.register(self)
        self.__dialog = dialog
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/PopoverJavaScript.ui")
        builder.connect_signals(self)
        widget = builder.get_object("widget")
        label = builder.get_object("label")
        image = builder.get_object("image")
        self.__entry = builder.get_object("entry")
        ok_button = builder.get_object("ok_button")
        cancel_button = builder.get_object("cancel_button")
        dialog_type = dialog.get_dialog_type()
        # Set icon
        if dialog_type == WebKit2.ScriptDialogType.ALERT:
            image.set_from_icon_name("dialog-warning-symbolic",
                                     Gtk.IconSize.DIALOG)
        elif dialog_type in [WebKit2.ScriptDialogType.CONFIRM,
                             WebKit2.ScriptDialogType.PROMPT,
                             WebKit2.ScriptDialogType.BEFORE_UNLOAD_CONFIRM]:
            image.set_from_icon_name("dialog-question-symbolic",
                                     Gtk.IconSize.DIALOG)
            ok_button.show()
            cancel_button.show()
        if dialog_type == WebKit2.ScriptDialogType.PROMPT:
            self.__entry.set_text(dialog.prompt_get_default_text())
            self.__entry.show()
        if dialog_type == WebKit2.ScriptDialogType.BEFORE_UNLOAD_CONFIRM:
            ok_button.set_label(_("Continue"))
            cancel_button.set_label(_("Cancel"))
        label.set_text(dialog.get_message())
        self.add(widget)
        self.__loop = GLib.MainLoop.new(None, False)
        self.connect("closed", self.__on_closed)

    def popup(self):
        """
            Popup widget and run loop
        """
        Gtk.Popover.popup(self)
        self.__loop.run()

#######################
# PROTECTED           #
#######################
    def _on_ok_button_clicked(self, button):
        """
            Pass ok to js
            @param button as Gtk.Button
        """
        if self.__entry.is_visible():
            self.__dialog.prompt_set_text(self.__entry.get_text())
        else:
            self.__dialog.confirm_set_confirmed(True)
        self.hide()

    def _on_cancel_button_clicked(self, button):
        """
            Pass ok to js
            @param button as Gtk.Button
        """
        if not self.__entry.is_visible():
            self.__dialog.confirm_set_confirmed(False)
        self.hide()

#######################
# PRIVATE             #
#######################
    def __on_closed(self, popover):
        """
            Quit main loop
            @param popover as Gtk.Popover
        """
        self.__loop.quit()
        self.__loop.unref()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Pango

from eolie.define import App


class ModifyUADialog:
    """
        Modify user agent for uri
    """

    def __init__(self, uri, window):
        """
            Init widget
            @param uri as str
            @param window as Gtk.Window
        """
        self.__uri = uri
        self.__window = window
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/DialogModifyUA.ui")
        builder.connect_signals(self)
        self.__dialog = builder.get_object("dialog")
        self.__dialog.set_transient_for(window)
        headerbar = builder.get_object("headerbar")
        view = builder.get_object("view")
        self.__model = builder.get_object("model")
        self.__selection = builder.get_object("selection")
        self.__entry = builder.get_object("entry")
        self.__entry.connect("changed", self.__on_entry_changed)
        self.__dialog.set_titlebar(headerbar)
        renderer0 = Gtk.CellRendererText()
        renderer0.set_property("ellipsize-set", True)
        renderer0.set_property("ellipsize", Pango.EllipsizeMode.END)
        column = Gtk.TreeViewColumn("")
        column.set_expand(True)
        column.pack_start(renderer0, True)
        column.add_attribute(renderer0, "text", 0)
        view.append_column(column)
        user_agent = App().websettings.get("user_agent", self.__uri)
        if user_agent is not None:
            self.__select_matching(user_agent)
            self.__entry.set_text(user_agent)

    def run(self):
        """
            Run dialog
        """
        self.__dialog.run()
        self.__dialog.destroy()

#######################
# PROTECTED           #
#######################
    def _on_selection_changed(self, selection):
        """
            Update entry
            @param selection as Gtk.TreeSelection
        """
        (model, items) = selection.get_selected_rows()
        if items:
            value = model[items[0]][1]
            # Can't set a null value in glade :-/
            if value == "None":
                value = ""
            self.__entry.disconnect_by_func(self.__on_entry_changed)
            self.__entry.set_text(value)
            self.__entry.connect("changed", self.__on_entry_changed)

    def _on_dialog_response(self, dialog, response_id):
        """
            Save user agent
            @param dialog as Gtk.Dialog
            @param response_id as int
        """
        if response_id != Gtk.ResponseType.DELETE_EVENT:
            user_agent = self.__entry.get_text()
            App().websettings.set("user_agent", self.__uri, user_agent)

#######################
# PRIVATE             #
#######################
    def __select_matching(self, user_agent):
        """
            Select matching element
            @param user_agent as str
        """
        for item in self.__model:
            value = "" if item[1] == "None" else item[1]
            if value == user_agent:
                self.__selection.select_iter(item.iter)
                return
        self.__selection.unselect_all()

    def __on_entry_changed(self, entry):
        """
            Update matching
            @param entry as Gtk.Entry
        """
        self.__select_matching(entry.get_text())



# Copyright (c) 2014-2020 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Gdk

from gettext import gettext as _

from eolie.utils import get_current_monitor_model
from eolie.helper_passwords import PasswordsHelper
from eolie.logger import Logger
from eolie.define import App


class SettingsDialog:
    """
        Dialog showing Eolie settings
    """

    __BOOLEAN = ["night-mode", "use-system-fonts", "remember-session",
                 "open-downloads", "enable-plugins", "developer-extras",
                 "do-not-track", "autoplay-videos", "remember-passwords",
                 "enable-firefox-sync", "enable-suggestions"]

    __RANGE = ["min-font-size", "max-popular-items"]
    __COMBO = ["start-page", "cookie-storage", "history-storage"]
    __ENTRY = ["start-page-custom"]
    __LOCKED_ON = {"use-system-fonts": ["sans-serif_row",
                                        "serif_row",
                                        "monospace_row"]}
    __LOCKED_OFF = {"enable-firefox-sync": ["status_row",
                                            "sync_row",
                                            "configure_row"]}

    def __init__(self, window):
        """
            Init dialog
            @param window as Window
        """
        self.__window = window
        self.__locked = []
        self.__passwords_helper = PasswordsHelper()
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/SettingsDialog.ui")
        self.__settings_dialog = builder.get_object("settings_dialog")
        self.__firefox_sync_button = builder.get_object(
            "enable-firefox-sync_boolean")
        self.__start_page_custom_entry = builder.get_object(
            "start-page-custom_entry")
        # Firefox sync
        self.__status_row = builder.get_object("status_row")
        self.__sync_button = builder.get_object("sync_button")
        self.__username_entry = builder.get_object("username_entry")
        self.__password_entry = builder.get_object("password_entry")
        self.__code_entry = builder.get_object("code_entry")
        for dic in [self.__LOCKED_ON, self.__LOCKED_OFF]:
            for key in dic.keys():
                for locked in dic[key]:
                    widget = builder.get_object(locked)
                    widget.set_name(key)
                    self.__locked.append(widget)
        self.__set_default_zoom_level(builder.get_object("default_zoom_level"))
        download_chooser = builder.get_object("download_chooser")
        dir_uri = App().settings.get_value("download-uri").get_string()
        if not dir_uri:
            directory = GLib.get_user_special_dir(
                GLib.UserDirectory.DIRECTORY_DOWNLOAD)
            if directory is not None:
                dir_uri = GLib.filename_to_uri(directory, None)
        if dir_uri:
            download_chooser.set_uri(dir_uri)
        else:
            download_chooser.set_uri("file://" + GLib.getenv("HOME"))
        script_uri = App().settings.get_value("user-script-uri").get_string()
        builder.get_object("user_script_chooser").set_uri(script_uri)
        builder.connect_signals(self)
        for setting in self.__BOOLEAN:
            button = builder.get_object("%s_boolean" % setting)
            value = App().settings.get_value(setting)
            # Only setting switch on fires a signal
            if value:
                button.set_state(value)
            else:
                self._on_boolean_state_set(button, False)
        for setting in self.__ENTRY:
            entry = builder.get_object("%s_entry" % setting)
            value = App().settings.get_value(setting).get_string()
            entry.set_text(value)
        for setting in self.__RANGE:
            widget = builder.get_object("%s_range" % setting)
            value = App().settings.get_value(setting).get_int32()
            widget.set_value(value)
        for setting in self.__COMBO:
            widget = builder.get_object("%s_combo" % setting)
            value = App().settings.get_enum(setting)
            widget.set_active(value)
        self.__multi_press = Gtk.EventControllerKey.new(self.__settings_dialog)
        self.__multi_press.connect("key-released", self.__on_key_released)
        self.__passwords_helper.get_sync(self.__on_get_sync)
        self.__check_sync_status()
        if App().sync_worker is not None:
            App().sync_worker.connect("syncing", self.__on_syncing)

    def show(self):
        """
            Show dialog
        """
        self.__settings_dialog.show()

    @property
    def stack(self):
        """
            Get main stack
            @return Gtk.Stack
        """
        return self.__settings_dialog.get_child()

#######################
# PROTECTED           #
#######################
    def _on_dialog_destroy(self, widget):
        """
            Clean up
            @param widget as Gtk.Widget
        """
        if App().sync_worker is not None:
            App().sync_worker.disconnect_by_func(self.__on_syncing)

    def _on_boolean_state_set(self, widget, state):
        """
            Save setting
            @param widget as Gtk.Switch
            @param state as bool
        """
        setting = widget.get_name()
        App().settings.set_value(setting,
                                 GLib.Variant("b", state))
        if setting in self.__LOCKED_ON.keys():
            self.__lock_for_setting(setting, not state)
        elif setting in self.__LOCKED_OFF.keys():
            self.__lock_for_setting(setting, state)
        if setting == "enable-firefox-sync" and\
                App().sync_worker is not None:
            if state:
                App().sync_worker.set_credentials()
                App().sync_worker.pull_loop()
                self.__status_row.set_subtitle(_("Connecting…"))
                GLib.timeout_add(5000, self.__check_sync_status)
            else:
                App().sync_worker.stop(True)
                self.__status_row.set_subtitle(_("Not running"))

    def _on_range_changed(self, widget):
        """
            Save value
            @param widget as Gtk.Range
        """
        setting = widget.get_name()
        value = widget.get_value()
        App().settings.set_value(setting, GLib.Variant("i", value))

    def _on_entry_changed(self, widget):
        """
            Save value
            @param widget as Gtk.Entry
        """
        setting = widget.get_name()
        value = widget.get_text()
        App().settings.set_value(setting, GLib.Variant("s", value))

    def _on_combo_changed(self, widget):
        """
            Save value
            @param widget as Gtk.ComboBoxText
        """
        setting = widget.get_name()
        value = widget.get_active()
        App().settings.set_enum(setting, value)
        if setting == "start-page":
            if value == 4:
                self.__start_page_custom_entry.show()
            else:
                self.__start_page_custom_entry.hide()

    def _on_download_selection_changed(self, chooser):
        """
            Save uri
            @chooser as Gtk.FileChooserButton
        """
        uri = chooser.get_uri()
        if uri is None:
            uri = ""
        App().settings.set_value("download-uri", GLib.Variant("s", uri))

    def _on_user_script_selection_changed(self, chooser):
        """
            Save uri
            @chooser as Gtk.FileChooserButton
        """
        uri = chooser.get_uri()
        if uri is None:
            uri = ""
        App().settings.set_value("user-script-uri", GLib.Variant("s", uri))

    def _on_font_sans_serif_set(self, fontbutton):
        """
            Save font setting
            @param fontchooser as Gtk.FontButton
        """
        value = GLib.Variant("s", fontbutton.get_font_name())
        App().settings.set_value("font-sans-serif", value)
        App().set_setting("sans-serif-font-family", fontbutton.get_font_name())

    def _on_font_serif_set(self, fontbutton):
        """
            Save font setting
            @param fontchooser as Gtk.FontButton
        """
        value = GLib.Variant("s", fontbutton.get_font_name())
        App().settings.set_value("font-serif", value)
        App().set_setting("serif-font-family", fontbutton.get_font_name())

    def _on_font_monospace_set(self, fontbutton):
        """
            Save font setting
            @param fontchooser as Gtk.FontButton
        """
        value = GLib.Variant("s", fontbutton.get_font_name())
        App().settings.set_value("font-monospace", value)
        App().set_setting("monospace-font-family", fontbutton.get_font_name())

    def _on_configure_engines_clicked(self, button):
        """
            Show Web engines configurator
            @param button as Gtk.Button
        """
        from eolie.dialog_search_engine import SearchEngineDialog
        dialog = SearchEngineDialog()
        dialog.show()
        dialog.connect("destroy-me", self.__on_sub_dialog_destroyed)
        self.stack.add(dialog)
        self.stack.set_visible_child(dialog)

    def _on_configure_notifications_clicked(self, button):
        """
            Show Web engines configurator
            @param button as Gtk.Button
        """
        from eolie.dialog_notifications import NotificationsDialog
        dialog = NotificationsDialog()
        dialog.show()
        dialog.connect("destroy-me", self.__on_sub_dialog_destroyed)
        self.stack.add(dialog)
        self.stack.set_visible_child(dialog)

    def _on_clear_personnal_data_clicked(self, button):
        """
            Show clear personnal data dialog
            @param button as Gtk.button
        """
        from eolie.dialog_clear_data import ClearDataDialog
        dialog = ClearDataDialog()
        dialog.show()
        dialog.connect("destroy-me", self.__on_sub_dialog_destroyed)
        self.stack.add(dialog)
        self.stack.set_visible_child(dialog)

    def _on_manage_passwords_clicked(self, button):
        """
            Launch searhorse
            @param button as Gtk.Button
        """
        from eolie.dialog_credentials import CredentialsDialog
        dialog = CredentialsDialog()
        dialog.show()
        dialog.connect("destroy-me", self.__on_sub_dialog_destroyed)
        self.stack.add(dialog)
        self.stack.set_visible_child(dialog)

    def _on_sync_now_clicked(self, button):
        """
            Sync now with Firefox Sync
            @param button as Gtk.Button
        """
        App().sync_worker.pull(True)
        App().sync_worker.push()

    def _on_sync_button_clicked(self, button):
        """
            Connect to Firefox Sync to get tokens
            @param button as Gtk.Button
        """
        if button.get_style_context().has_class("suggested-action"):
            button.get_style_context().remove_class("destructive-action")
            self.__status_row.set_subtitle(_("Connecting…"))
            App().task_helper.run(self.__connect_firefox_sync,
                                  self.__username_entry.get_text(),
                                  self.__password_entry.get_text(),
                                  self.__code_entry.get_text(),
                                  callback=(self.__on_sync_result,))
        else:
            App().sync_worker.stop(True)
            App().sync_worker.delete_secret()
            button.get_style_context().remove_class("suggested-action")
            button.get_style_context().add_class("destructive-action")
            self.__status_row.set_subtitle(_("Stopped…"))

    def _on_credentials_changed(self, entry):
        """
            Update widgets state
            @param entry as Gtk.Entry
        """
        credentials_ok = self.__username_entry.get_text() != "" and (
            self.__password_entry.get_text() != "" or
            self.__code_entry.get_text() != "")
        self.__sync_button.set_sensitive(credentials_ok)
        if self.__password_entry.get_text() != "":
            self.__password_entry.set_sensitive(True)
            self.__code_entry.set_sensitive(False)
        elif self.__code_entry.get_text() != "":
            self.__password_entry.set_sensitive(False)
            self.__code_entry.set_sensitive(True)
        else:
            self.__password_entry.set_sensitive(True)
            self.__code_entry.set_sensitive(True)

#######################
# PRIVATE             #
#######################
    def __get_sync_status(self):
        """
            Get sync status
            return int
        """
        return App().sync_worker.status

    def __check_sync_status(self):
        """
            Check worker status
        """
        if App().sync_worker is not None:
            App().task_helper.run(self.__get_sync_status,
                                  callback=(self.__on_get_sync_status,))
        else:
            App().settings.set_value("enable-firefox-sync",
                                     GLib.Variant("b", False))
            self.__firefox_sync_button.set_sensitive(False)
            from eolie.firefox_sync import SyncWorker
            if True or not SyncWorker.check_modules():
                cmd =\
                 "$ pip3 install requests-hawk PyFxA pycrypto cryptography"
                self.__status_row.set_subtitle(cmd)
            else:
                self.__status_row.set_subtitle(_("Not running"))

    def __lock_for_setting(self, setting, sensitive):
        """
            Lock widgets for setting
            @param setting as str
            @param sensitive as bool
        """
        for locked in self.__locked:
            if locked.get_name() == setting:
                locked.set_sensitive(sensitive)

    def __connect_firefox_sync(self, username, password, code):
        """
            Connect to firefox sync
            @param username as str
            @param password as str
            @param code as str
            @return bool
        """
        try:
            App().sync_worker.new_session()
            App().sync_worker.login({"login": username}, password, code)
            return True
        except Exception as e:
            Logger.error("SettingsDialog::__connect_firefox_sync(): %s", e)
            if str(e) == "Unverified account":
                return -1
        return False

    def __set_default_zoom_level(self, widget):
        """
            Set default zoom level
            @param widget as Gtk.SpinButton
        """
        monitor_model = get_current_monitor_model(self.__window)
        zoom_levels = App().settings.get_value("default-zoom-level")
        wanted_zoom_level = 1.0
        try:
            for zoom_level in zoom_levels:
                zoom_splited = zoom_level.split('@')
                if zoom_splited[0] == monitor_model:
                    wanted_zoom_level = float(zoom_splited[1])
        except:
            pass
        percent_zoom = int(wanted_zoom_level * 100)
        widget.set_value(percent_zoom)
        widget.set_text("{} %".format(percent_zoom))
        widget.connect("value-changed", self.__on_default_zoom_changed)

    def __on_default_zoom_changed(self, button):
        """
            Save size
            @param button as Gtk.SpinButton
        """
        button.set_text("{} %".format(int(button.get_value())))
        monitor_model = get_current_monitor_model(self.__window)
        try:
            # Add current value less monitor model
            zoom_levels = []
            for zoom_level in App().settings.get_value("default-zoom-level"):
                zoom_splited = zoom_level.split('@')
                if zoom_splited[0] == monitor_model:
                    continue
                else:
                    zoom_levels.append("%s@%s" % (zoom_splited[0],
                                                  zoom_splited[1]))
            # Add new zoom value for monitor model
            zoom_levels.append("%s@%s" % (monitor_model,
                                          button.get_value() / 100))
            App().settings.set_value("default-zoom-level",
                                     GLib.Variant("as", zoom_levels))
            for window in App().windows:
                window.update_zoom_level(True)
        except Exception as e:
            Logger.error("SettingsDialog::__on_default_zoom_changed(): %s", e)

    def __on_key_released(self, event_controller, keyval, keycode, state):
        """
            Quit on escape
            @param event_controller as Gtk.EventController
            @param keyval as int
            @param keycode as int
            @param state as Gdk.ModifierType
        """
        if keyval == Gdk.KEY_Escape:
            self.__settings_dialog.destroy()

    def __on_sync_result(self, result):
        """
            Show result to user
            @param result as bool
        """
        if result is True:
            self.__status_row.set_subtitle(_("Connected"))
        elif result == -1:
            self.__status_row.set_subtitle(
                _("Check your email and connect again"))
        else:
            self.__status_row.set_subtitle(_("Failed"))

    def __on_get_sync(self, attributes, password, uri, index, count):
        """
            Set username and password
            @param attributes as {}
            @param password as str
            @param uri as None
            @param index as int
            @param count as int
        """
        if attributes is None:
            return
        try:
            self.__username_entry.set_text(attributes["login"])
        except Exception as e:
            Logger.error("SettingsDialog::__on_get_sync(): %s", e)

    def __on_get_sync_status(self, status):
        """
            Show a message about missing fxa module
            @param status as bool
        """
        if status:
            self.__status_row.set_subtitle(_("Connected"))
            App().sync_worker.pull_loop()

    def __on_syncing(self, worker, message):
        """
            Show message as status
            @param worker as SyncWorker
            @param message as str
        """
        self.__status_row.set_subtitle(_("Syncing %s") % message)

    def __on_sub_dialog_destroyed(self, widget):
        """
            Restore previous dialog
            @param widget as Gtk.Widget
        """
        for child in self.stack.get_children():
            if child != widget:
                self.stack.set_visible_child(child)
                break
        GLib.timeout_add(1000, widget.destroy)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, Gtk, GLib

from gettext import gettext as _

from eolie.define import App, LoadingType


class MoveToMenu(Gtk.Grid):
    """
        Menu allowing to move webviews to a window
    """

    def __init__(self, webviews, window, back=True):
        """
            Init menu
            @param webviews as [WebView]
            @param window as Window
            @param back as bool
        """
        self.__window = window
        self.__actions = []
        self.__webviews = list(webviews)
        Gtk.Menu.__init__(self)
        self.set_margin_start(5)
        self.set_margin_end(5)
        self.set_margin_top(5)
        self.set_margin_bottom(5)
        self.set_orientation(Gtk.Orientation.VERTICAL)

        # Back button
        if back:
            item = Gtk.ModelButton.new()
            item.set_hexpand(True)
            item.set_property("centered", True)
            item.set_property("text", _("Move to"))
            item.set_property("inverted", True)
            item.set_property("menu-name", "main")
            item.show()
            self.add(item)

        action = Gio.SimpleAction(name="switch_window")
        action = Gio.SimpleAction.new("switch_window",
                                      GLib.VariantType.new("s"))
        action.connect('activate',
                       self.__on_action_activate)
        window.add_action(action)

        # New window button
        item = Gtk.ModelButton.new()
        item.set_hexpand(True)
        item.set_property("text", _("New window"))
        item.set_action_name("win.switch_window")
        item.set_action_target_value(GLib.Variant("s", "new_window"))
        item.show()
        self.add(item)
        if len(App().windows) > 1:
            item = Gtk.Separator.new(Gtk.Orientation.HORIZONTAL)
            item.show()
            self.add(item)

        for _window in App().windows:
            if _window == window:
                continue
            item = Gtk.ModelButton.new()
            item.set_hexpand(True)
            item.set_property("text", _window.get_title())
            item.set_action_name("win.switch_window")
            item.set_action_target_value(GLib.Variant("s", str(_window)))
            item.show()
            item.connect("enter-notify-event",
                         self.__on_enter_notify_event,
                         _window)
            item.connect("leave-notify-event",
                         self.__on_leave_notify_event,
                         _window)
            self.add(item)

    def do_hide(self):
        """
            Remove action on hide
        """
        Gtk.Grid.do_hide(self)
        self.__window.remove_action("switch_window")

#######################
# PRIVATE             #
#######################
    def __on_enter_notify_event(self, widget, event, window):
        """
            Mark window
            @param widget as Gtk.Widget
            @param event as Gdk.Event
            @param window as Window
        """
        window.mark(True)

    def __on_leave_notify_event(self, widget, event, window):
        """
            Unmark window
            @param widget as Gtk.Widget
            @param event as Gdk.Event
            @param window as Window
        """
        window.mark(False)

    def __on_action_activate(self, action, variant):
        """
            Moves webviews to window
            @param action as  Gio.SimpleAction
            @param variant as GLib.Variant
            @param window as Window
        """
        window = None
        window_str = variant.get_string()

        # Get wanted window
        if window_str == "new_window":
            window = App().get_new_window()
        else:
            for window in App().windows:
                if window_str == str(window):
                    break
        if window is None:
            return

        # Move webviews to window
        for webview in self.__webviews:
            self.__window.container.remove_webview(webview)
            window.container.add_webview(webview, LoadingType.FOREGROUND)
            webview.set_window(window)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio

from hashlib import sha256
from gettext import gettext as _

from eolie.define import App


class FormMenu(Gio.Menu):
    """
        Menu showing form username
    """

    def __init__(self, webview, window):
        """
            Init menu
            @param webview as Webview
            @param window as Window
        """
        Gio.Menu.__init__(self)
        self.__window = window
        self.__webview = webview
        self.__section = Gio.Menu()
        self.append_section(_("Saved credentials"), self.__section)

    def add_attributes(self, attributes, uri):
        """
            Add username to model
            @param attributes as {}
            @param uri as str
        """
        encoded = "FORM_" + sha256(
            attributes["login"].encode("utf-8")).hexdigest()
        action = Gio.SimpleAction(name=encoded)
        App().add_action(action)
        action.connect('activate',
                       self.__on_action_clicked,
                       attributes["uuid"])
        label = attributes["login"].replace("_", "__")
        item = Gio.MenuItem.new(label, "app.%s" % encoded)
        self.__section.append_item(item)

#######################
# PRIVATE             #
#######################
    def __on_action_clicked(self, action, variant, uuid):
        """
            Update form
            @param Gio.SimpleAction
            @param GVariant
            @param uuid as str
        """
        self.__webview.set_forms_content(uuid)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import WebKit2, Gio, GLib

from time import time

from eolie.define import App
from eolie.css_stylesheets import StyleSheets


class WebViewNightMode:
    """
        Night mode for webview
    """

    def __init__(self):
        """
            Init night mode
        """
        self.__night_mode = False
        self.__started_time = 0
        self.__css_uri = None
        self.__cancellable = Gio.Cancellable.new()
        self.__stylesheets = StyleSheets()
        self.__stylesheets.set_cancellable(self.__cancellable)
        self.__stylesheets.connect("not-cached",
                                   self.__on_stylesheets_not_cached)
        self.__stylesheets.connect("populated",
                                   self.__on_stylesheets_populated)
        self.get_style_context().add_class("night-mode")
        self.__default_stylesheet = WebKit2.UserStyleSheet(
                     "body, table, figure {\
                        color: #EAEAEA !important;\
                        background-color: #353535 !important\
                      }\
                      a {\
                        color: #F0FFFE !important;\
                      }\
                      a:hover {\
                        color: white !important;\
                      }\
                      p[style] {\
                           background-color: #303030 !important\
                      }\
                      input, select, button{\
                        color: #EAEAEA !important;\
                        background-color: #353535 !important;\
                        border: 1px solid grey;\
                      }",
                     WebKit2.UserContentInjectedFrames.ALL_FRAMES,
                     WebKit2.UserStyleLevel.USER,
                     None,
                     None)

    def load_css_uri(self, message):
        """
            Load CSS URI as user style
            @param message as str
        """
        self.__css_uri = self.uri
        self.__stylesheets.load_css_uri(message, self.__started_time)

    def load_css_text(self, message):
        """
            Load CSS text as user style
            @param message as str
        """
        self.__css_uri = self.uri
        self.__stylesheets.load_css_text(message,
                                         self.uri,
                                         self.__started_time)

    def night_mode(self):
        """
            Handle night mode
        """
        self.__stylesheets.reset()
        self.__css_uri = None
        if self.__should_apply_night_mode():
            self.run_javascript_from_gresource(
                "/org/gnome/Eolie/javascript/GetCSS.js", None, None)
        else:
            self.get_user_content_manager().remove_all_style_sheets()

    def remove_night_mode_cache(self):
        """
            Remove cache for current stylesheets
        """
        self.__stylesheets.remove_cache()

    @property
    def stylesheets(self):
        """
            Get stylesheets object
            @return StyleSheets
        """
        return self.__stylesheets

#######################
# PROTECTED           #
#######################
    def _on_load_changed(self, webview, event):
        """
            Run JS helpers
            @param webview as WebView
            @param event as WebKit2.LoadEvent
        """
        if not self.__should_apply_night_mode():
            self.get_user_content_manager().remove_all_style_sheets()
            return
        if event == WebKit2.LoadEvent.STARTED:
            self.__css_uri = None
            self.__started_time = int(time())
            self.__cancellable.cancel()
            self.__cancellable = Gio.Cancellable.new()
            self.__stylesheets.set_cancellable(self.__cancellable)
        elif event == WebKit2.LoadEvent.REDIRECTED:
            self.__css_uri = None
        elif webview.uri != self.__css_uri:
            self.run_javascript_from_gresource(
                "/org/gnome/Eolie/javascript/GetCSS.js", None, None)

#######################
# PRIVATE             #
#######################
    def __should_apply_night_mode(self):
        """
            True if night mode should be applied
            @return bool
        """
        night_mode = App().settings.get_value("night-mode")
        netloc_night_mode = App().websettings.get("night_mode", self.uri)
        return night_mode and netloc_night_mode in [1, None]

    def __on_stylesheets_not_cached(self, stylesheets):
        """
            Apply stylesheets
            @param stylesheets as StyleSheets
        """
        self.run_javascript("""
                    html = document.querySelector("html");
                    if (html !== null) {
                        html.style.opacity = 0;
                    }""", None, None)

    def __on_stylesheets_populated(self, stylesheets):
        """
            Apply stylesheets
            @param stylesheets as StyleSheets
        """
        content_manager = self.get_user_content_manager()
        content_manager.remove_all_style_sheets()
        content_manager.add_style_sheet(self.__default_stylesheet)
        user_style_sheet = WebKit2.UserStyleSheet(
                 stylesheets.get_css_text(self.__started_time),
                 WebKit2.UserContentInjectedFrames.ALL_FRAMES,
                 WebKit2.UserStyleLevel.USER,
                 None,
                 None)
        content_manager.add_style_sheet(user_style_sheet)
        GLib.timeout_add(250, self.run_javascript, """
            html = document.querySelector("html");
            if (html !== null) {
                html.style.opacity = 1;
            }""", None, None)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GObject, Pango, GLib

from eolie.helper_passwords import PasswordsHelper
from eolie.define import App, MARGIN_SMALL
from eolie.utils import emit_signal
from eolie.logger import Logger


class Row(Gtk.ListBoxRow):
    """
        A row
    """

    def __init__(self, scheme, host):
        """
            Init Row
            @param scheme as str
            @param host as str
        """
        Gtk.ListBoxRow.__init__(self)
        self.__scheme = scheme
        self.__host = host
        label = Gtk.Label.new("%s://%s" % (scheme, host))
        label.show()
        label.set_halign(Gtk.Align.START)
        label.set_property("margin", MARGIN_SMALL)
        label.set_ellipsize(Pango.EllipsizeMode.END)
        self.add(label)

    @property
    def scheme(self):
        """
            Get scheme
            @return str
        """
        return self.__scheme

    @property
    def host(self):
        """
            Get host
            @return str
        """
        return self.__host


class NotificationsDialog(Gtk.Bin):
    """
        Show notifications authorizations
    """

    __gsignals__ = {
        "destroy-me": (GObject.SignalFlags.RUN_FIRST, None, ())
    }

    def __init__(self):
        """
            Init dialog
        """
        Gtk.Bin.__init__(self)
        self.__filter = ""
        self.__helper = PasswordsHelper()
        builder = Gtk.Builder()
        builder.add_from_resource('/org/gnome/Eolie/DialogNotifications.ui')
        builder.connect_signals(self)
        self.__search_bar = builder.get_object("search_bar")
        self.__remove_button = builder.get_object("remove_button")
        self.__listbox = builder.get_object("listbox")
        self.__listbox.set_filter_func(self.__filter_func)
        self.__listbox.set_sort_func(self.__sort_func)
        self.add(builder.get_object('widget'))
        self.__populate()

#######################
# PROTECTED           #
#######################
    def _on_back_clicked(self, button):
        """
            Ask to be destroyed
            @param button as Gtk.Button
        """
        emit_signal(self, "destroy-me")

    def _on_search_changed(self, entry):
        """
            Update filter
            @param entry as Gtk.Entry
        """
        self.__filter = entry.get_text()
        self.__listbox.invalidate_filter()

    def _on_remove_clicked(self, button):
        """
            Remove all passwords
            @param button as Gtk.Button
        """
        values = list(App().settings.get_value("notification-domains"))
        for row in self.__listbox.get_selected_rows():
            values.remove("%s;%s" % (row.scheme, row.host))
            row.destroy()
        App().settings.set_value("notification-domains",
                                 GLib.Variant("as", values))

    def _on_row_selected(self, listbox, row):
        """
            Update clear button state
            @param listbox as Gtk.ListBox
            @param row as Gtk.ListBoxRow
        """
        self.__remove_button.set_sensitive(
            len(listbox.get_selected_rows()) != 0)

    def _on_search_toggled(self, button):
        """
            Show entry
            @param button as Gtk.Button
        """
        self.__search_bar.set_search_mode(button.get_active())

#######################
# PRIVATE             #
#######################
    def __filter_func(self, row):
        """
            Filter rows
            @param row as Row
        """
        return self.__filter in row.host

    def __sort_func(self, row1, row2):
        """
            Sort rows
            @param row1 as Row
            @param row2 as Row
        """
        return row2.host < row1.host

    def __populate(self):
        """
            Populate view
        """
        values = App().settings.get_value("notification-domains")
        try:
            for value in values:
                (scheme, host) = value.split(";")
                row = Row(scheme, host)
                row.show()
                self.__listbox.add(row)
        except Exception as e:
            Logger.error("NotificationsDialog::__populate(): %s", e)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib, Gtk

from gettext import gettext as _

from eolie.define import App


class SitesMenu(Gtk.Grid):
    """
        Menu linked to a SitesManagerChild
    """

    def __init__(self, webviews, window):
        """
            Init menu
            @param webviews as [WebView]
            @param window as Window
        """
        self.__window = window
        self.__webviews = webviews
        Gtk.Grid.__init__(self)
        self.set_margin_start(5)
        self.set_margin_end(5)
        self.set_margin_top(5)
        self.set_margin_bottom(5)
        self.set_orientation(Gtk.Orientation.VERTICAL)
        # Page switcher
        action = Gio.SimpleAction.new("switch_page",
                                      GLib.VariantType.new("s"))
        self.__window.add_action(action)
        action.connect("activate",
                       self.__on_action_activate)
        for webview in webviews:
            uri = webview.uri
            if uri is None:
                continue
            title = webview.title
            item = Gtk.ModelButton.new()
            item.set_hexpand(True)
            item.set_property("text", title)
            item.set_action_name("win.switch_page")
            item.set_action_target_value(GLib.Variant("s", str(webview)))
            item.show()
            self.add(item)
        # Bottom section
        separator = Gtk.Separator.new(Gtk.Orientation.HORIZONTAL)
        separator.show()
        self.add(separator)
        # Move to
        item = Gtk.ModelButton.new()
        item.set_property("text", _("Move to"))
        item.set_property("menu-name", "moveto")
        item.show()
        self.add(item)
        # Modify UA
        if App().settings.get_value("developer-extras"):
            action = Gio.SimpleAction.new("user_agent")
            action.connect("activate", self.__on_modify_ua_activate)
            self.__window.add_action(action)
            item = Gtk.ModelButton.new()
            item.set_property("text", _("Modify user agent"))
            item.set_action_name("win.user_agent")
            item.show()
            self.add(item)
        # Pinned state
        uri = webviews[0].uri
        pinned = App().websettings.get("pinned", uri)
        action = Gio.SimpleAction.new_stateful(
                    "pinned",
                    None,
                    GLib.Variant.new_boolean(pinned))
        action.connect("change-state",
                       self.__on_pinned_change_state,
                       uri)
        self.__window.add_action(action)
        item = Gtk.ModelButton.new()
        item.set_property("text", _("Pinned site"))
        item.set_action_name("win.pinned")
        item.show()
        self.add(item)
        # Close site
        action = Gio.SimpleAction.new("close_site")
        action.connect("activate", self.__on_close_activate)
        self.__window.add_action(action)
        item = Gtk.ModelButton.new()
        item.set_property("text", _("Close site"))
        item.set_action_name("win.close_site")
        item.show()
        self.add(item)

#######################
# PRIVATE             #
#######################
    def __on_pinned_change_state(self, action, param, uri):
        """
            Set option value
            @param action as Gio.SimpleAction
            @param param as GLib.Variant
            @param uri as str
        """
        action.set_state(param)
        App().websettings.set("pinned", uri, param.get_boolean())

    def __on_close_activate(self, action, param):
        """
            Close wanted page
            @param action as Gio.SimpleAction
            @param param as GLib.Variant
        """
        for webview in self.__webviews:
            self.__window.container.try_close_webview(webview)

    def __on_modify_ua_activate(self, action, param):
        """
            Show a dialog allowing user to update User Agent
            @param action as Gio.SimpleAction
            @param param as GLib.Variant
        """
        if self.__webviews:
            from eolie.dialog_modify_ua import ModifyUADialog
            dialog = ModifyUADialog(self.__webviews[0].uri, self.__window)
            dialog.run()

    def __on_action_activate(self, action, variant):
        """
            Switch view
            @param action as Gio.SimpleAction
            @param variant as GLib.Variant
            @param view as View
        """
        webview_str = variant.get_string()
        for webview in self.__window.container.webviews:
            if webview_str == str(webview):
                self.__window.container.set_visible_webview(webview)
                break



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib

import sqlite3
import itertools
from urllib.parse import urlparse
from threading import Lock

from eolie.sqlcursor import SqlCursor
from eolie.define import EOLIE_DATA_PATH, Type
from eolie.logger import Logger
from eolie.database_upgrade import DatabaseUpgrade
from eolie.utils import get_safe_netloc


class DatabaseSettings:
    """
        Store various settings for webpage
    """

    # Allows to return someting if not result
    __DEFAULTS = {"audio": 1}
    # SQLite documentation:
    # In SQLite, a column with type INTEGER PRIMARY KEY
    # is an alias for the ROWID.
    # Here, we define an id INT PRIMARY KEY but never feed it,
    # this make VACUUM not destroy rowids...
    __create_settings = '''CREATE TABLE settings (
                                           id INTEGER PRIMARY KEY,
                                           netloc TEXT NOT NULL,
                                           chooser_uri TEXT,
                                           languages TEXT,
                                           zoom INT,
                                           accept_tls INT NOT NULL DEFAULT 0,
                                           night_mode INT,
                                           pinned INT NOT NULL DEFAULT 0,
                                           geolocation INT,
                                           user_agent TEXT,
                                           audio INT NOT NULL DEFAULT 1
                                           )'''

    def __init__(self):
        """
            Create database tables or manage update if needed
            @param suffix as str
        """
        self.thread_lock = Lock()
        self.__DB_PATH = "%s/settings.db" % EOLIE_DATA_PATH
        upgrade = DatabaseUpgrade(Type.SETTINGS)
        if not GLib.file_test(self.__DB_PATH, GLib.FileTest.IS_REGULAR):
            try:
                if not GLib.file_test(EOLIE_DATA_PATH, GLib.FileTest.IS_DIR):
                    GLib.mkdir_with_parents(EOLIE_DATA_PATH, 0o0750)
                # Create db schema
                with SqlCursor(self, True) as sql:
                    sql.execute(self.__create_settings)
                    sql.execute("PRAGMA user_version=%s" % upgrade.version)
            except Exception as e:
                Logger.error("DatabaseSettings::__init__(): %s", e)
        else:
            upgrade.upgrade(self)

    def set(self, option, uri, status):
        """
            Set option for URI to value
            @param option as str
            @param uri as str
            @param status as object
        """
        parsed = urlparse(uri)
        if parsed.scheme not in ["http", "https"]:
            return
        try:
            netloc = get_safe_netloc(uri)
            with SqlCursor(self, True) as sql:
                result = sql.execute("SELECT rowid FROM settings\
                                      WHERE netloc=?", (netloc,))
                v = result.fetchone()
                if v is not None:
                    sql.execute("UPDATE settings\
                                 SET %s=?\
                                 WHERE netloc=?" % option,
                                (status, netloc))
                else:
                    sql.execute("INSERT INTO settings\
                                          (netloc, %s)\
                                          VALUES (?, ?)" % option,
                                (netloc, status))
        except Exception as e:
            Logger.error("DatabaseSettings::set(): %s", e)

    def get(self, option, uri):
        """
            Get option for URI
            @param option as str
            @param uri as str
            @return object
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT %s FROM settings\
                                  WHERE netloc=?" % option,
                                 (get_safe_netloc(uri),))
            v = result.fetchone()
            if v is not None:
                return v[0]
            if option in self.__DEFAULTS.keys():
                return self.__DEFAULTS[option]
            return None

    def get_languages(self, uri):
        """
            Get languages for uri
            @param uri as str
            @return codes as [str]
            @raise if not found
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT languages FROM settings\
                                  WHERE netloc=?", (get_safe_netloc(uri),))
            v = result.fetchone()
            if v is not None:
                languages = v[0]
                if languages:
                    return languages.split(";")
                else:
                    return []
            else:
                return None

    def get_pinned_netlocs(self):
        """
            Get pinned netlocs
            return [str]
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT netloc FROM settings WHERE pinned=1")
            return list(itertools.chain(*result))

    def add_language(self, code, uri):
        """
            Add language for uri
            @param code as str
            @param uri as str
        """
        parsed = urlparse(uri)
        if parsed.scheme not in ["http", "https"]:
            return
        try:
            with SqlCursor(self, True) as sql:
                codes = self.get_languages(uri)
                if codes is not None:
                    if code not in codes:
                        codes.append(code)
                    sql.execute("UPDATE settings\
                                 SET languages=?\
                                 WHERE netloc=?", (";".join(codes),
                                                   get_safe_netloc(uri)))
                else:
                    sql.execute("INSERT INTO settings\
                                          (uri, languages)\
                                          VALUES (?, ?)",
                                (get_safe_netloc(uri), code))
        except Exception as e:
            Logger.error("DatabaseSettings::add_language(): %s", e)

    def remove_language(self, code, uri):
        """
            Remove language for uri
            @param code as str
            @param uri as str
        """
        codes = self.get_languages(uri)
        if codes is not None and code in codes:
            codes.remove(code)
            with SqlCursor(self, True) as sql:
                sql.execute("UPDATE settings\
                                 SET languages=?\
                                 WHERE netloc=?", (";".join(codes),
                                                   get_safe_netloc(uri)))

    def get_cursor(self):
        """
            Return a new sqlite cursor
        """
        try:
            c = sqlite3.connect(self.__DB_PATH, 600.0)
            return c
        except Exception as e:
            Logger.error("DatabaseSettings::get_cursor(): %s", e)
            exit(-1)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Handy

from eolie.toolbar_actions import ToolbarActions
from eolie.toolbar_title import ToolbarTitle
from eolie.toolbar_end import ToolbarEnd
from eolie.helper_gestures import GesturesHelper


class Toolbar(Handy.HeaderBar):
    """
        Eolie toolbar
    """

    def __init__(self, window, fullscreen=False):
        """
            Init toolbar
            @param window as Window
            @param fullscreen as bool
        """
        Handy.HeaderBar.__init__(self)
        self.__window = window
        self.set_title("Eolie")
        self.__toolbar_actions = ToolbarActions(window, fullscreen)
        self.__toolbar_actions.show()
        self.__toolbar_title = ToolbarTitle(window)
        self.__toolbar_title.show()
        self.__toolbar_end = ToolbarEnd(window, fullscreen)
        self.__toolbar_end.show()
        self.pack_start(self.__toolbar_actions)
        self.set_custom_title(self.__toolbar_title)
        self.pack_end(self.__toolbar_end)
        self.__gesture = GesturesHelper(
            self,
            primary_press_callback=self.__on_press)

    @property
    def headerbar(self):
        """
            Get headerbar
            @return Gtk.HeaderBar
        """
        return self.__headerbar

    @property
    def title(self):
        """
            Toolbar title
            @return ToolbarTitle
        """
        return self.__toolbar_title

    @property
    def end(self):
        """
            Toolbar end
            @return ToolbarEnd
        """
        return self.__toolbar_end

    @property
    def actions(self):
        """
            Toolbar actions
            @return ToolbarActions
        """
        return self.__toolbar_actions

#######################
# PRIVATE             #
#######################
    def __on_press(self, x, y, event):
        """
            Hide popovers
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        self.__window.close_popovers()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Gio, Gdk, Handy

from eolie.define import App, LoadingType
from eolie.toolbar import Toolbar
from eolie.container import Container
from eolie.utils import get_current_monitor_model, emit_signal
from eolie.helper_task import TaskHelper
from eolie.logger import Logger
from eolie.window_state import WindowState


class Window(Handy.ApplicationWindow, WindowState):
    """
        Main window
    """

    def __init__(self, size, is_maximized):
        """
            Init window
            @param app as Gtk.Application
            @param size as (int, int)
            @param is_maximized as bool
        """
        Handy.ApplicationWindow.__init__(self,
                                         application=App(),
                                         title="Eolie",
                                         icon_name="org.gnome.Eolie")
        WindowState.__init__(self)
        self.__monitor_model = ""
        self.__popovers = []
        self.__zoom_level = 1.0
        self.__container = None
        self.__modifiers = 0
        self.__window_state = 0
        self.__timeout_configure = None
        self.__size = size
        self.set_default_size(size[0], size[1])
        self.__setup_content()
        self.connect("realize", self.__on_realize, is_maximized)
        self.connect("window-state-event", self.__on_window_state_event)
        self.connect("configure-event", self.__on_configure_event)
        self.__key_event_controller = Gtk.EventControllerKey.new(self)
        self.__key_event_controller.connect("key-pressed",
                                            self.__on_key_pressed)
        self.__key_event_controller.connect("key-released",
                                            self.__on_key_released)

        # Set window actions
        shortcut_action = Gio.SimpleAction.new("shortcut",
                                               GLib.VariantType.new("s"))
        shortcut_action.connect("activate", self.__on_shortcut_action)
        self.add_action(shortcut_action)
        self.set_auto_startup_notification(False)

    def update_zoom_level(self, force):
        """
            Update zoom level
            @param force as bool
        """
        try:
            monitor_model = get_current_monitor_model(self)
            if force or monitor_model != self.__monitor_model:
                # Update window default zoom level
                self.__update_zoom_level()
                self.__monitor_model = monitor_model
                # Update view zoom level
                for webview in self.__container.webviews:
                    webview.update_zoom_level()
        except Exception as e:
            Logger.error("Window::update_zoom_level(): %s", e)

    def fullscreen(self, force=True):
        """
            Prepare window to fullscreen and enter fullscreen
            @param force as bool
        """
        if self.__fullscreen_revealer is not None:
            return
        self.__fullscreen_toolbar = Toolbar(self, True)
        # Do not count container.webviews as destroy may be pending on somes
        count = str(len(self.__container.pages_manager.children))
        self.__fullscreen_toolbar.actions.count_label.set_text(count)
        self.__fullscreen_toolbar.show()
        self.__fullscreen_revealer = Gtk.Revealer.new()
        self.__fullscreen_revealer.set_property("valign", Gtk.Align.START)
        self.__fullscreen_revealer.add(self.__fullscreen_toolbar)
        self.__fullscreen_revealer.show()
        self.__container.overlay.add_overlay(self.__fullscreen_revealer)
        self.__container.sites_manager.hide()
        self.__toolbar.hide()
        if force:
            Gtk.ApplicationWindow.fullscreen(self)
        self.connect("motion-notify-event", self.__on_motion_notify_event)

    def unfullscreen(self, force=True):
        """
            Prepare window to unfullscreen and leave fullscreen
            @param force as bool
        """
        if self.__fullscreen_revealer is None:
            return
        self.disconnect_by_func(self.__on_motion_notify_event)
        GLib.idle_add(self.__fullscreen_toolbar.destroy)
        GLib.idle_add(self.__fullscreen_revealer.destroy)
        if App().settings.get_value("show-sidebar"):
            self.__container.sites_manager.show()
        self.__fullscreen_toolbar = None
        self.__fullscreen_revealer = None
        self.__toolbar.show()
        # Do not count container.webviews as destroy may be pending on somes
        # Reason: we do not remove/destroy view to let stack animation run
        count = len(self.container.pages_manager.children)
        self.toolbar.actions.count_label.set_text(str(count))
        if force:
            Gtk.ApplicationWindow.unfullscreen(self)
        self.__container.webview.run_javascript(
            "document.webkitExitFullscreen();",
            None,
            None)

    def register(self, popover, monitor=True):
        """
            Add a popover to window
            @param popover as Gtk.Popover
            @param monitor as bool, check if closed
        """
        self.__popovers.append(popover)
        if monitor:
            popover.connect("closed", self.__on_popover_closed)

    def close_popovers(self):
        """
            Close all popovers
            @return closed as bool
        """
        closed = False
        for popover in self.__popovers:
            if popover.is_visible():
                closed = True
            popover.popdown()
        return closed

    def mark(self, mark):
        """
            Change window toolbar color
        """
        style = self.__toolbar.get_style_context()
        if mark:
            style.add_class("toolbar-mark")
        else:
            style.remove_class("toolbar-mark")

    @property
    def container(self):
        """
            Get window container
            @return Container
        """
        return self.__container

    @property
    def monitor_model(self):
        """
            Get current monitor model
        """
        self.__monitor_model

    @property
    def toolbar(self):
        """
            Get window toolbar
            @return Toolbar
        """
        if self.__fullscreen_toolbar is None:
            return self.__toolbar
        else:
            return self.__fullscreen_toolbar

    @property
    def zoom_level(self):
        """
           Get zoom level for window
           @return float
        """
        return self.__zoom_level

    @property
    def is_fullscreen(self):
        """
            True if fullscreen
            @return bool
        """
        return self.__window_state & Gdk.WindowState.FULLSCREEN

    @property
    def size(self):
        """
            Unis_maximized window size
            return (int, int)
        """
        return self.__size

############
# Private  #
############
    def __update_zoom_level(self):
        """
            Update zoom level for default screen
        """
        monitor_model = get_current_monitor_model(self)
        zoom_levels = App().settings.get_value(
            "default-zoom-level")
        user_zoom_level = False
        for zoom_level in zoom_levels:
            zoom_splited = zoom_level.split("@")
            if zoom_splited[0] == monitor_model:
                self.__zoom_level = float(zoom_splited[1])
                user_zoom_level = True
                break
        if not user_zoom_level:
            self.__zoom_level = 1.0

    def __setup_content(self):
        """
            Setup window content
        """
        grid = Gtk.Grid()
        grid.set_orientation(Gtk.Orientation.VERTICAL)
        grid.show()
        self.__toolbar = Toolbar(self)
        self.__fullscreen_toolbar = None
        self.__fullscreen_revealer = None
        self.__toolbar.show()
        self.__container = Container(self)
        self.__container.show()
        self.__toolbar.set_show_close_button(True)
        grid.add(self.__toolbar)
        grid.add(self.__container)
        self.add(grid)
        self.get_style_context().connect(
            "changed", lambda x: self.__toolbar.title.entry.update_style())

    def __setup_window(self, is_maximized):
        """
            Set window
            @param is_maximized as bool
        """
        try:
            if is_maximized:
                self.maximize()
        except Exception as e:
            Logger.error("Window::__setup_window(): %s", e)

    def __on_configure_event(self, window, event):
        """
            Delay event
            @param: window as Gtk.Window
            @param: event as Gdk.Event
        """
        size = window.get_size()
        self.toolbar.title.set_width(size[0] / 5)
        if self.__timeout_configure:
            GLib.source_remove(self.__timeout_configure)
            self.__timeout_configure = None
        self.__timeout_configure = GLib.timeout_add(
            250,
            self.__on_configure_timeout)

    def __on_source_loaded(self, uri, status, content):
        """
            Show source code for uri
            @param uri as str
            @param status as bool
            @param content as bytes
        """
        try:
            (tmp, tmp_stream) = Gio.File.new_tmp("XXXXXX.html")
            tmp_stream.get_output_stream().write_all(content)
            tmp_stream.close()
            appinfo = Gio.app_info_get_default_for_type("text/plain", False)
            appinfo.launch([tmp], None)
        except Exception as e:
            Logger.error("Window::__on_source_loaded(): %s", e)

    def __on_configure_timeout(self):
        """
            Update zoom level
        """
        self.__timeout_configure = None
        self.update_zoom_level(False)
        if not self.is_maximized():
            self.__size = self.get_size()

    def __on_window_state_event(self, widget, event):
        """
            Save maximised state
            @param: window as Gtk.Window
            @param: event as Gdk.EventWindowState
        """
        if event.changed_mask & Gdk.WindowState.MAXIMIZED:
            size = widget.get_size()
            self.toolbar.title.set_width(size[0] / 5)
        self.__window_state = event.new_window_state

    def __on_motion_notify_event(self, widget, event):
        """
            Reaveal/hide toolbar if needed
            @param: widget as Gtk.Widget
            @param: event as Gdk.EventMotion
        """
        if self.__fullscreen_revealer.get_reveal_child() and\
                not self.__fullscreen_revealer.get_child_revealed():
            return
        if event.y < self.__fullscreen_revealer.get_allocated_height():
            self.__fullscreen_revealer.set_reveal_child(True)
        else:
            lock = False
            for popover in self.__popovers:
                if popover.is_visible():
                    lock = True
                    break
            if not lock:
                self.__fullscreen_revealer.set_reveal_child(False)

    def __on_realize(self, widget, is_maximized):
        """
            Update zoom level
            @param widget as Gtk.Widget
            @param is_maximized as bool
        """
        self.__setup_window(is_maximized)
        self.update_zoom_level(False)

    def __on_key_pressed(self, controller, keyval, keycode, state):
        """
            Update PagesManager sort on Ctrl<Tab>
            @param controller as Gtk.EventControllerKey
            @param keyval as int
            @param keycode as int
            @param state as Gtk.ModifierType
            @param event as Gdk.EventKey
        """
        if state & Gdk.ModifierType.CONTROL_MASK:
            if keyval == Gdk.KEY_Tab and not self.container.in_expose:
                self.container.pages_manager.update_sort()
            elif keyval == Gdk.KEY_z:
                self.container.webview.run_javascript_from_gresource(
                    "/org/gnome/Eolie/javascript/HandleInputPrev.js",
                    None, None)
            elif keyval == Gdk.KEY_Z:
                self.container.webview.run_javascript_from_gresource(
                    "/org/gnome/Eolie/javascript/HandleInputNext.js",
                    None, None)

    def __on_key_released(self, controller, keyval, keycode, state):
        """
            Handle Esc/Ctrl release
            @param controller as Gtk.EventControllerKey
            @param keyval as int
            @param keycode as int
            @param state as Gtk.ModifierType
            @param event as Gdk.EventKey
        """
        if keyval == Gdk.KEY_Control_L:
            self.__container.ctrl_released()
        elif keyval == Gdk.KEY_Escape:
            self.__container.set_expose(False)
            if self.__container.reading:
                self.__container.toggle_reading()

    def __on_shortcut_action(self, action, param):
        """
            Global shortcuts handler
            @param action as Gio.SimpleAction
            @param param as GLib.Variant
        """
        string = param.get_string()
        if string == "uri":
            if self.is_fullscreen and self.__fullscreen_revealer is not None:
                self.__fullscreen_revealer.set_reveal_child(True)
            self.toolbar.title.entry.focus()
        elif string == "fullscreen":
            if self.is_fullscreen:
                self.unfullscreen()
            else:
                self.fullscreen()
        elif string == "quit":
            App().quit(True)
        elif string == "new_page":
            self.container.add_webview_for_uri(App().start_page,
                                               LoadingType.FOREGROUND)
            self.toolbar.title.start_search()
        elif string == "close_page":
            if self.is_fullscreen:
                emit_signal(self.container.webview, "leave-fullscreen")
                Gtk.ApplicationWindow.unfullscreen(self)
            self.__container.try_close_webview(self.container.webview)
        elif string == "reload":
            self.container.webview.reload()
        elif string == "settings":
            # Rework all this code to use actions like in Lollypop
            from eolie.dialog_settings import SettingsDialog
            dialog = SettingsDialog(self)
            dialog.show()
        elif string == "home":
            self.container.webview.load_uri(App().start_page)
        elif string == "source":
            uri = self.container.webview.uri
            task_helper = TaskHelper()
            task_helper.load_uri_content(uri, None, self.__on_source_loaded)
        elif string == "find":
            find_widget = self.container.find_widget
            find_widget.set_search_mode(True)
            find_widget.search()
        elif string == "backward":
            self.toolbar.actions.backward()
        elif string == "forward":
            self.toolbar.actions.forward()
        elif string == "previous":
            self.__container.previous()
        elif string == "next":
            self.__container.next()
        elif string == "previous_site":
            self.__container.sites_manager.previous()
        elif string == "next_site":
            self.__container.sites_manager.next()
        elif string == "print":
            self.container.webview.print()
        elif string == "private":
            self.container.add_webview_for_uri(App().start_page,
                                               LoadingType.FOREGROUND,
                                               True)
        elif string == "last_page":
            App().pages_menu.activate_last_action()
        elif string == "zoom_in":
            self.container.webview.zoom_in()
        elif string == "zoom_out":
            self.container.webview.zoom_out()
        elif string == "zoom_default":
            self.container.webview.zoom_default()
        elif string == "history":
            self.toolbar.title.entry.focus("history")
        elif string == "search":
            self.toolbar.title.entry.focus("search")
        elif string == "save":
            self.toolbar.end.save_page()
        elif string == "expose":
            active = self.toolbar.actions.view_button.get_active()
            self.toolbar.actions.view_button.set_active(not active)
        elif string == "show_sidebar":
            value = App().settings.get_value("show-sidebar")
            App().settings.set_value("show-sidebar",
                                     GLib.Variant("b", not value))
        elif string == "mse_enabled":
            self.container.webview.set_setting("enable-mediasource", True)
            self.container.webview.reload()

    def __on_popover_closed(self, popover):
        """
            Remove popover from registered
            @param popover as Gtk.Popover
        """
        if popover in self.__popovers:
            self.__popovers.remove(popover)
            GLib.timeout_add(1000, popover.destroy)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from gettext import gettext as _

from eolie.define import App, MARGIN_SMALL


class UriEntryIcons(Gtk.Grid):
    """
        Entry for main URI bar
    """

    def __init__(self, entry, window):
        """
            Init toolbar
            @param entry as Gtk.Entry
            @param window as Window
        """
        Gtk.Grid.__init__(self)
        self.set_margin_end(2)
        self.set_margin_start(2)
        self.__entry = entry
        self.__window = window

        separator = Gtk.Separator.new(Gtk.Orientation.VERTICAL)
        separator.show()
        separator.set_size_request(-1, 32)
        separator.set_margin_top(MARGIN_SMALL)
        separator.set_margin_bottom(MARGIN_SMALL)

        self.__reload_stop_button = Gtk.Button.new_from_icon_name(
            "view-refresh-symbolic", Gtk.IconSize.BUTTON)
        self.__reload_stop_button.show()
        self.__reload_stop_button.get_style_context().add_class(
            "no-border-button")
        self.__reload_stop_button.connect("clicked",
                                          self.__on_reload_button_clicked)

        self.__bookmark_button = Gtk.Button.new_from_icon_name(
            "non-starred-symbolic", Gtk.IconSize.BUTTON)
        self.__bookmark_button.show()
        self.__bookmark_button.get_style_context().add_class(
            "no-border-button")
        self.__bookmark_button.connect("clicked",
                                       self.__on_bookmark_button_clicked)

        self.__spinner = Gtk.Spinner.new()
        self.__spinner.show()
        # Magic value to be aligned with buttons
        self.__spinner.set_margin_end(7)

        self.__readability_button = Gtk.Button.new_from_icon_name(
            "view-dual-symbolic", Gtk.IconSize.BUTTON)
        self.__readability_button.show()
        self.__readability_button.get_style_context().add_class(
            "no-border-button")
        self.__readability_button.connect("clicked",
                                          self.__on_readability_button_clicked)

        self.__geolocation_button = Gtk.Button.new_from_icon_name(
            "mark-location-symbolic", Gtk.IconSize.BUTTON)
        self.__geolocation_button.get_style_context().add_class(
            "no-border-button")
        self.__geolocation_button.set_tooltip_text(_("Deny geolocation"))
        self.__geolocation_button.connect("clicked",
                                          self.__on_geolocation_button_clicked)

        self.add(self.__spinner)
        self.add(self.__geolocation_button)
        self.add(self.__readability_button)
        self.add(self.__bookmark_button)
        self.add(separator)
        self.add(self.__reload_stop_button)

    def set_readable_button_state(self, reading):
        """
            Mark readable button
            @param reading as bool
        """
        image = self.__readability_button.get_image()
        style_context = image.get_style_context()
        if reading:
            style_context.add_class("selected")
        else:
            style_context.remove_class("selected")

    def set_loading(self, loading):
        """
            Mark current webview as loading
            @param loading as bool
        """
        image = self.__reload_stop_button.get_image()
        if loading:
            self.__spinner.start()
            self.__spinner.show()
            image.set_from_icon_name(
                "process-stop-symbolic", Gtk.IconSize.BUTTON)
        else:
            self.__spinner.stop()
            self.__spinner.hide()
            image.set_from_icon_name(
                "view-refresh-symbolic", Gtk.IconSize.BUTTON)

    def set_bookmarked(self, bookmarked):
        """
            Set current webview as bookmarked
            @param bookmarked as bool
        """
        image = self.__bookmark_button.get_image()
        if bookmarked:
            image.set_from_icon_name(
                "starred-symbolic", Gtk.IconSize.BUTTON)
        else:
            image.set_from_icon_name(
                "non-starred-symbolic", Gtk.IconSize.BUTTON)

    def show_readable_button(self, show):
        """
            Show readable button
            @param show as bool
        """
        if show:
            self.__readability_button.show()
            self.set_readable_button_state(False)
        else:
            self.__readability_button.hide()

    def show_geolocation(self, show):
        """
            Show geolocation button
            @param show as bool
        """
        if show:
            self.__geolocation_button.show()
        else:
            self.__geolocation_button.hide()

    def show_clear_button(self):
        """
            Show clear button
        """
        if self.__window.container.webview.uri != "about:blank":
            image = self.__bookmark_button.get_image()
            image.set_from_icon_name("edit-clear-symbolic",
                                     Gtk.IconSize.BUTTON)

#######################
# PRIVATE             #
#######################
    def __on_readability_button_clicked(self, button):
        """
            Switch reading mode
            @param button as Gtk.Button
        """
        reading = self.__window.container.toggle_reading()
        self.set_readable_button_state(reading)

    def __on_geolocation_button_clicked(self, button):
        """
            Disable geolocation for current
            @param button as Gtk.Button
        """
        uri = self.__window.container.webview.uri
        App().websettings.set("geolocation", uri, False)

    def __on_reload_button_clicked(self, button):
        """
            Reload current view/Stop loading
            @param button as Gtk.Button
        """
        icon_name = self.__reload_stop_button.get_image().get_icon_name()[0]
        if icon_name == 'view-refresh-symbolic':
            self.__window.container.webview.reload_bypass_cache()
        else:
            self.__window.container.webview.stop_loading()

    def __on_bookmark_button_clicked(self, button):
        """
            Add/Remove page to/from bookmarks
            @param button as Gtk.Button
        """
        def on_popover_closed(popover, webview):
            bookmark_id = App().bookmarks.get_id(webview.uri)
            image = self.__bookmark_button.get_image()
            if bookmark_id is None:
                image.set_from_icon_name(
                    "non-starred-symbolic", Gtk.IconSize.BUTTON)
            else:
                image.set_from_icon_name(
                    "starred-symbolic", Gtk.IconSize.BUTTON)

        webview = self.__window.container.webview
        icon_name = self.__bookmark_button.get_image().get_icon_name()[0]
        if icon_name == "edit-clear-symbolic":
            self.__entry.delete_text(0, -1)
            webview.clear_text_entry()
            self.__entry.grab_focus()
        else:
            bookmark_id = App().bookmarks.get_id(webview.uri)
            if bookmark_id is None:
                image = self.__bookmark_button.get_image()
                image.set_from_icon_name(
                    "starred-symbolic", Gtk.IconSize.BUTTON)
                bookmark_id = App().bookmarks.add(webview.title,
                                                  webview.uri, None, [])
            from eolie.widget_bookmark_edit import BookmarkEditWidget
            widget = BookmarkEditWidget(bookmark_id, False)
            widget.show()
            popover = Gtk.Popover.new()
            popover.set_relative_to(button)
            popover.connect("closed", on_popover_closed, webview)
            popover.add(widget)
            popover.popup()



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


class WindowStateStruct:
    def __init__(self):
        self.wid = ""
        self.size = (800, 600)
        self.is_maximized = False
        self.webview_states = []
        self.sort = []


class WindowState:
    """
        Window state allowing to restore a webview from disk
    """

    def new_from_state(state):
        """
            New webview from WindowStateStruct
            @param state as WebState
        """
        from eolie.window import Window
        window = Window(state.size, state.is_maximized)
        window.container.sites_manager.set_initial_sort(state.sort)
        return window

    def __init__(self):
        """
            Init state
        """
        pass

    @property
    def state(self):
        """
            Get state
            @return WindowStateStruct
        """
        state = WindowStateStruct()
        for webview in self.container.webviews:
            webview_state = webview.state
            if webview_state is not None:
                state.webview_states.append(webview_state)
        state.size = self.size
        state.is_maximized = self.is_maximized()
        state.sort = self.container.sites_manager.sort
        return state



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, GObject, GLib, Pango

from gettext import gettext as _
from datetime import datetime

from eolie.define import App, Type, LoadingType, MARGIN_SMALL
from eolie.utils import wanted_loading_type, emit_signal


class Row(Gtk.ListBoxRow):
    """
        A row
    """
    __gsignals__ = {
        'edited': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'moved': (GObject.SignalFlags.RUN_FIRST, None, (GLib.Variant,))
    }

    def __init__(self, item, window):
        """
            Init row
            @param item as Item
            @param window as Window
        """
        self.__item = item
        self.__window = window
        self.__search = ""
        eventbox = None
        favicon = None
        Gtk.ListBoxRow.__init__(self)
        item_id = item.get_property("id")
        item_type = item.get_property("type")
        uri = item.get_property("uri")
        title = item.get_property("title")
        grid = Gtk.Grid()
        grid.set_property("margin", MARGIN_SMALL)
        grid.set_column_spacing(10)
        grid.set_hexpand(True)
        grid.set_property("valign", Gtk.Align.CENTER)
        if item_type in [Type.BOOKMARK, Type.SEARCH, Type.HISTORY]:
            favicon = Gtk.Image()
            self.__set_favicon(favicon)
        elif item_type == Type.SUGGESTION:
            favicon = Gtk.Image.new_from_icon_name("system-search-symbolic",
                                                   Gtk.IconSize.MENU)
            favicon.show()
        elif item_type == Type.WEBVIEW:
            favicon = Gtk.Image.new_from_icon_name("view-paged-symbolic",
                                                   Gtk.IconSize.MENU)
            favicon.show()
        if favicon is not None:
            grid.attach(favicon, 0, 0, 1, 2)

        uri = item.get_property("uri")
        self.__title = Gtk.Label.new()
        self.__title.set_ellipsize(Pango.EllipsizeMode.END)
        self.__title.set_property("halign", Gtk.Align.START)
        self.__title.set_hexpand(True)
        self.__title.set_property('has-tooltip', True)
        self.__title.connect('query-tooltip', self.__on_query_tooltip)
        self.__title.show()
        if uri:
            self.__title.set_markup("%s\n<span alpha='40000'>%s</span>" %
                                    (GLib.markup_escape_text(title),
                                     GLib.markup_escape_text(uri)))
        else:
            self.__title.set_text(title)
        grid.attach(self.__title, 1, 0, 1, 2)

        if item_type == Type.HISTORY:
            dt = datetime.fromtimestamp(item.get_property("atime"))
            hour = str(dt.hour).rjust(2, "0")
            minute = str(dt.minute).rjust(2, "0")
            atime = Gtk.Label.new("%s:%s" % (hour, minute))
            atime.set_property("valign", Gtk.Align.CENTER)
            atime.get_style_context().add_class("dim-label")
            atime.set_margin_end(2)
            atime.show()
            grid.attach(atime, 2, 0, 1, 2)

        if item_type == Type.HISTORY:
            delete_button = Gtk.Button.new_from_icon_name(
                "user-trash-symbolic",
                Gtk.IconSize.MENU)
            delete_button.get_image().set_opacity(0.5)
            delete_button.set_margin_end(5)
            delete_button.set_property("valign", Gtk.Align.CENTER)
            delete_button.connect("clicked", self.__on_delete_clicked)
            delete_button.get_style_context().add_class("overlay-button")
            delete_button.set_tooltip_text(_("Delete page from history"))
            delete_button.show()
            grid.attach(delete_button, 3, 0, 1, 2)
        elif item_type == Type.BOOKMARK:
            edit_button = Gtk.Button.new_from_icon_name(
                "document-edit-symbolic",
                Gtk.IconSize.MENU)
            edit_button.get_image().set_opacity(0.5)
            edit_button.set_margin_end(5)
            edit_button.connect("clicked", self.__on_edit_clicked)
            edit_button.get_style_context().add_class("overlay-button")
            edit_button.set_property("valign", Gtk.Align.CENTER)
            edit_button.set_tooltip_text(_("Edit bookmark"))
            edit_button.show()
            grid.attach(edit_button, 2, 0, 1, 2)
        elif item_type == Type.TAG:
            if item_id == Type.NONE:
                icon_name = "folder-visiting-symbolic"
            elif item_id == Type.POPULARS:
                icon_name = "emote-love-symbolic"
            elif item_id == Type.RECENTS:
                icon_name = "document-open-recent-symbolic"
            else:
                icon_name = "folder-symbolic"
            open_button = Gtk.Button.new_from_icon_name(
                icon_name,
                Gtk.IconSize.MENU)
            open_button.connect("clicked", self.__on_open_clicked)
            open_button.get_style_context().add_class("overlay-button-alt")
            open_button.set_tooltip_text(_("Open all pages with this tag"))
            open_button.show()
            grid.attach(open_button, 0, 0, 1, 1)
        grid.show()
        style_context = self.get_style_context()
        style_context.add_class("row")
        eventbox = Gtk.EventBox()
        eventbox.add(grid)
        eventbox.set_size_request(-1, 30)
        eventbox.connect("button-release-event",
                         self.__on_button_release_event)
        eventbox.show()
        self.add(eventbox)

    def set_title(self, title):
        """
            Set row title
            @param title as str
        """
        self.__item.set_property("title", title)
        self.__title.set_text(title)

    @property
    def item(self):
        """
            Get item
            @return Item
        """
        return self.__item

#######################
# PRIVATE             #
#######################
    def __set_favicon(self, favicon):
        """
            Try to get a favicon for current URI
            @param favicon as Gtk.Image
            @param uri as str
        """
        uri = self.__item.get_property("uri")
        favicon_path = App().art.get_favicon_path(uri)
        if favicon_path is not None:
            favicon.set_from_file(favicon_path)
        else:
            favicon.set_from_icon_name("web-browser-symbolic",
                                       Gtk.IconSize.LARGE_TOOLBAR)
        favicon.show()

    def __on_query_tooltip(self, widget, x, y, keyboard, tooltip):
        """
            Show tooltip if needed
            @param widget as Gtk.Widget
            @param x as int
            @param y as int
            @param keyboard as bool
            @param tooltip as Gtk.Tooltip
        """
        text = ''
        layout = widget.get_layout()
        label = widget.get_text()
        if layout.is_ellipsized():
            text = "%s" % (GLib.markup_escape_text(label))
        widget.set_tooltip_markup(text)

    def __on_button_release_event(self, eventbox, event):
        """
            Handle button press in popover
            @param eventbox as Gtk.EventBox
            @param event as Gdk.Event
        """
        # Lets user select item
        if event.state & Gdk.ModifierType.CONTROL_MASK or\
                event.state & Gdk.ModifierType.SHIFT_MASK:
            return False
        # Event is for internal button
        if eventbox.get_window() != event.window:
            return True
        uri = self.__item.get_property("uri")
        type_id = self.__item.get_property("type")
        if type_id in [Type.HISTORY, Type.SUGGESTION,
                       Type.SEARCH, Type.BOOKMARK]:
            if event.button == 1:
                self.__window.container.webview.load_uri(uri)
                self.__window.container.set_expose(False)
                self.__window.close_popovers()
            else:
                self.__window.container.add_webview_for_uri(
                    uri, LoadingType.FOREGROUND)
                if event.button == 2:
                    self.__window.close_popovers()
        elif type_id == Type.WEBVIEW:
            title = self.__item.get_property("title")
            for webview in self.__window.container.webviews:
                if webview.uri == uri and webview.title == title:
                    self.__window.container.set_visible_webview(webview)
                    self.__window.close_popovers()
                    break
        else:
            emit_signal(self, "activate")
            # We force focus to stay on title entry
            GLib.idle_add(self.__window.toolbar.title.entry.focus)

    def __on_edit_clicked(self, button):
        """
            Edit self
            @param button as Gtk.Button
        """
        emit_signal(self, "edited")

    def __on_open_clicked(self, button):
        """
            Open all bookmarks
            @param button as Gtk.Button
        """
        self.__window.close_popovers()
        tag_id = self.__item.get_property("id")
        if tag_id == Type.POPULARS:
            items = App().bookmarks.get_populars(50)
        elif tag_id == Type.RECENTS:
            items = App().bookmarks.get_recents()
        elif tag_id == Type.UNCLASSIFIED:
            items = App().bookmarks.get_unclassified()
        else:
            items = App().bookmarks.get_bookmarks(tag_id)
        i = 0
        for (bid, uri, title) in items:
            loading_type = wanted_loading_type(i)
            self.__window.container.add_webview_for_uri(uri, loading_type)
            i += 1

    def __on_delete_clicked(self, button):
        """
            Delete self
            @param button as Gtk.Button
        """
        history_id = self.__item.get_property("id")
        guid = App().history.get_guid(history_id)
        if App().sync_worker is not None:
            App().sync_worker.remove_from_history(guid)
        App().history.remove(history_id)
        GLib.idle_add(self.destroy)



# Copyright (c) 2017-2018 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# Copyright (c) 2015 Jean-Philippe Braun <eon@patapon.info>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from threading import current_thread

from eolie.define import App


class SqlCursor:
    """
        Context manager to get the SQL cursor
    """
    def add(obj):
        """
            Add cursor to thread list
        """
        name = current_thread().getName() + obj.__class__.__name__
        App().cursors[name] = obj.get_cursor()

    def remove(obj):
        """
            Remove cursor from thread list and commit
        """
        name = current_thread().getName() + obj.__class__.__name__
        if name in App().cursors.keys():
            obj.thread_lock.acquire()
            App().cursors[name].commit()
            obj.thread_lock.release()
            App().cursors[name].close()
            del App().cursors[name]

    def commit(obj):
        """
            Commit current obj
        """
        name = current_thread().getName() + obj.__class__.__name__
        if name in App().cursors.keys():
            obj.thread_lock.acquire()
            App().cursors[name].commit()
            obj.thread_lock.release()

    def __init__(self, obj, commit=False):
        """
            Init object
            @param obj as Database/Playlists/Radios
            @param commit as bool
        """
        self.__obj = obj
        self.__commit = commit
        self.__cursor = None

    def __enter__(self):
        """
            Get thread cursor or a new one
        """
        name = current_thread().getName() + self.__obj.__class__.__name__
        if name in App().cursors.keys():
            cursor = App().cursors[name]
            return cursor
        else:
            self.__cursor = self.__obj.get_cursor()
            return self.__cursor

    def __exit__(self, type, value, traceback):
        """
            Close cursor if not thread cursor
        """
        if self.__cursor is not None:
            if self.__commit:
                self.__obj.thread_lock.acquire()
                self.__cursor.commit()
                self.__obj.thread_lock.release()
            self.__cursor.close()
        self.__cursor = None



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Gdk

from gettext import gettext as _
from random import randint
from time import time

from eolie.popover_webview import WebViewPopover
from eolie.define import App
from eolie.container_sidebar import SidebarContainer
from eolie.container_expose import ExposeContainer
from eolie.container_stack import StackContainer
from eolie.container_overlay import OverlayContainer
from eolie.container_webview import WebViewContainer
from eolie.container_reading import ReadingContainer


class Container(Gtk.Paned,
                OverlayContainer, StackContainer,
                SidebarContainer, ExposeContainer,
                WebViewContainer, ReadingContainer):
    """
        Main Eolie view
    """

    __DONATION = 1

    def __init__(self, window):
        """
            Ini.container
            @param window as Window
        """
        self._window = window
        Gtk.Paned.__init__(self)
        StackContainer.__init__(self)
        OverlayContainer.__init__(self)
        SidebarContainer.__init__(self)
        ExposeContainer.__init__(self)
        WebViewContainer.__init__(self)
        ReadingContainer.__init__(self)
        self.__popover = WebViewPopover(window)
        # Show donation notification after one hour
        if App().settings.get_value("donation").get_int32() != self.__DONATION:
            GLib.timeout_add_seconds(randint(3600, 7200),
                                     self.__show_donation)

    def load_uri(self, uri):
        """
            Load uri in current view
            @param uri as str
        """
        if self.webview is not None:
            self.webview.load_uri(uri)

    def popup_webview(self, webview):
        """
            Show webview in popopver
            @param webview as WebView
        """
        self.__popover.add_webview(webview)
        if not self.__popover.is_visible():
            self.__popover.set_relative_to(self._window.toolbar)
            self.__popover.set_position(Gtk.PositionType.BOTTOM)
            self.__popover.popup()

    def set_visible_webview(self, webview):
        """
            Set visible webview
            @param webview as WebView
        """
        webview.set_shown(True)
        webview.set_atime(int(time()))
        self.sites_manager.update_shown_state(webview)
        self.pages_manager.update_shown_state(webview)
        StackContainer.set_visible_webview(self, webview)
        WebViewContainer.set_visible_webview(self, webview)
        ReadingContainer.set_visible_webview(self, webview)

#######################
# PRIVATE             #
#######################
    def __show_donation(self):
        """
            Show a notification telling user to donate a little
        """
        from eolie.container_notification import ContainerNotification
        notification = ContainerNotification(
            _("Please consider a donation to the project"),
            [_("PayPal"), _("Patreon")],
            [lambda: Gtk.show_uri_on_window(
                App().active_window,
                "https://www.paypal.me/lollypopgnome",
                Gdk.CURRENT_TIME),
             lambda: Gtk.show_uri_on_window(
                App().active_window,
                "https://www.patreon.com/gnumdk",
                Gdk.CURRENT_TIME)])
        self.overlay.add_overlay(notification)
        notification.show()
        notification.set_reveal_child(True)
        App().settings.set_value("donation",
                                 GLib.Variant("i", self.__DONATION))



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio

from gettext import gettext as _
from base64 import b64decode

from eolie.utils import emit_signal
from eolie.webview import WebView
from eolie.logger import Logger


class ReadingContainer:
    """
        Reading management for container
    """

    def __init__(self):
        """
            Init container
        """
        self._reading_webview = None
        self.__related_webview = None
        self.__related_webview_id = None

    def toggle_reading(self):
        """
            Toggle reading mode
            @return status as bool
        """
        if self._reading_webview is None:
            js1 = Gio.File.new_for_uri(
                "resource:///org/gnome/Eolie/Readability.js")
            js2 = Gio.File.new_for_uri(
                "resource:///org/gnome/Eolie/Readability_get.js")
            (status, content1, tags) = js1.load_contents()
            (status, content2, tags) = js2.load_contents()
            script = content1.decode("utf-8") + content2.decode("utf-8")
            self.webview.run_javascript(script, None,
                                        self.__on_readability_content)
            self.__related_webview = self.webview
            return True
        else:
            if self.__related_webview_id is not None:
                self.__related_webview.disconnect(self.__related_webview_id)
            self.__related_webview = None
            self.__related_webview_id = None
            self._reading_webview.destroy()
            self._reading_webview = None
            return False

    def set_visible_webview(self, webview):
        """
            Set visible webview
            @param webview as WebView
        """
        if self.reading:
            self.toggle_reading()
        emit_signal(webview, "readability-status", webview.readability_status)
        webview.check_readability()

    @property
    def reading(self):
        """
            True if reading
            @return bool
        """
        return self._reading_webview is not None

#######################
# PRIVATE             #
#######################
    def __on_decide_policy(self, webview, decision, decision_type):
        """
            Load link inside main view
            @param webview as WebKit2.WebView
            @param decision as WebKit2.NavigationPolicyDecision
            @param decision_type as WebKit2.PolicyDecisionType
            @return bool
        """
        navigation_action = decision.get_navigation_action()
        navigation_uri = navigation_action.get_request().get_uri()
        if navigation_uri == "about:blank":  # load_html()
            decision.use()
            return True
        else:
            self.webview.load_uri(navigation_uri)
            webview.destroy()
            return False

    def __on_readability_content(self, webview, result):
        """
            Show reading content
            @param webview as WebView
            @param result as Gio.AsyncResult
        """
        try:
            data = webview.run_javascript_from_gresource_finish(result)
            bytes = data.get_js_value().to_string_as_bytes()
            content = b64decode(bytes.get_data()).decode("utf-8")
        except Exception as e:
            Logger.error("ReadingContainer::__on_readability_content(): %s", e)
            content = _("Nothing to display")
        system = Gio.Settings.new("org.gnome.desktop.interface")
        document_font_name = system.get_value("document-font-name").get_string(
        )
        document_font_size = str(int(document_font_name[-2:]) * 1.3) + "pt"
        if self._reading_webview is None:
            self._reading_webview = WebView.new(webview.window)
            self._reading_webview.set_zoom_level(webview.get_zoom_level())
            self._reading_webview.connect("decide-policy",
                                          self.__on_decide_policy)
            self._reading_webview.show()
            self.overlay.add_overlay(self._reading_webview)
            html = "<html><head>\
                    <style type='text/css'>\
                    *:not(img) {font-size: %s;\
                        background-color: #333333;\
                        color: #e6e6e6;\
                        margin-left: auto;\
                        margin-right: auto;\
                        width: %s}\
                    </style></head>" % (document_font_size,
                                        self.get_allocated_width() / 1.5)
            html += "<title>%s</title>" % webview.title
            html += content
            html += "</html>"
            self._reading_webview.load_html(html, None)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from eolie.define import App, LoadingType
from eolie.utils import update_popover_internals
from eolie.helper_gestures import GesturesHelper


class ToolbarActions(Gtk.Bin):
    """
        Actions toolbar
    """

    def __init__(self, window, fullscreen):
        """
            Init toolbar
            @param window as Window
            @param fullscreen as bool
        """
        Gtk.Bin.__init__(self)
        self.__window = window
        self.__timeout_id = None
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Eolie/ToolbarActions.ui")
        builder.connect_signals(self)

        self.add(builder.get_object("actions"))
        self.set_hexpand(True)
        self.__backward_button = builder.get_object("back_button")
        self.__forward_button = builder.get_object("forward_button")
        self.__pages_button = builder.get_object("pages_button")
        self.__view_button = builder.get_object("view_button")
        self.__close_button = builder.get_object("close_button")
        self.__count = builder.get_object("count")

        self.__gesture1 = GesturesHelper(
            self.__backward_button,
            primary_press_callback=self.__on_back_primary_press,
            secondary_press_callback=self.__on_back_secondary_press,
            primary_long_callback=self.__on_back_primary_long)
        self.__gesture2 = GesturesHelper(
            self.__forward_button,
            primary_press_callback=self.__on_forward_primary_press,
            secondary_press_callback=self.__on_forward_secondary_press,
            primary_long_callback=self.__on_forward_primary_long)

    def set_actions(self, webview):
        """
            Set available actions based on view
            @param webview as WebView
        """
        self.__backward_button.set_sensitive(webview.can_go_back())
        self.__forward_button.set_sensitive(webview.can_go_forward())

    def backward(self):
        """
            Click next
        """
        webview = self.__window.container.webview
        webview.go_back()

    def forward(self):
        """
            Click previous
        """
        webview = self.__window.container.webview
        webview.go_forward()

    @property
    def count_label(self):
        """
            Get count label
            @return Gtk.Label
        """
        return self.__count

    @property
    def view_button(self):
        """
            Get view pages button
            @return Gtk.MenuButton
        """
        return self.__view_button

#######################
# PROTECTED           #
#######################
    def _on_new_button_clicked(self, button):
        """
            Add a new web view
            @param button as Gtk.Button
        """
        self.__window.container.add_webview_for_uri(App().start_page,
                                                    LoadingType.FOREGROUND)
        self.__window.close_popovers()
        self.__window.toolbar.title.start_search()

    def _on_application_button_toggled(self, button):
        """
            Show pages popover
            @param button as Gtk.ToggleButton
        """
        self.__window.close_popovers()
        if not button.get_active():
            return
        from eolie.menu_application import ApplicationMenu
        menu = ApplicationMenu(self.__window)
        popover = Gtk.Popover.new_from_model(button, menu)
        popover.set_modal(False)
        self.__window.register(popover)
        popover.connect("closed", lambda x: button.set_active(False))
        popover.popup()

    def _on_view_button_toggled(self, button):
        """
            Show current views
            @param button as Gtk.ToggleButton
        """
        active = button.get_active()
        self.__window.container.set_expose(active)
        self.__window.close_popovers()

    def _on_close_button_clicked(self, button):
        """
            Close current page
            @param button as button
        """
        self.__window.container.try_close_webview(
            self.__window.container.webview)

#######################
# PRIVATE             #
#######################
    def __show_menu(self, button):
        """
            Show back history
            @param button as Gtk.Button
        """
        current = self.__window.container.webview
        if button == self.__backward_button:
            history_list = current.get_back_forward_list().get_back_list()
        else:
            history_list = current.get_back_forward_list().get_forward_list()
        if history_list:
            from eolie.menu_history import HistoryMenu
            model = HistoryMenu(history_list, self.__window)
            popover = Gtk.Popover.new_from_model(button, model)
            popover.set_modal(False)
            self.__window.register(popover)
            GLib.idle_add(popover.forall, update_popover_internals)
            popover.connect("closed",
                            self.__on_navigation_popover_closed,
                            model)
            popover.popup()

    def __on_navigation_popover_closed(self, popover, model):
        """
            Clear menu actions
            @param popover
            @param model as HistoryMenu/None
        """
        # Let model activate actions, idle needed to action activate
        GLib.idle_add(model.clean)

    def __on_back_primary_press(self, x, y, event):
        """
            Go backward in history
            @param x as int
            @param y as int
            @param even as Gdk.Event
        """
        self.backward()

    def __on_back_secondary_press(self, x, y):
        """
            Show backward menu
            @param x as int
            @param y as int
        """
        self.__show_menu(self.__backward_button)

    def __on_back_primary_long(self, x, y):
        """
            Show backward menu
            @param x as int
            @param y as int
        """
        self.__show_menu(self.__backward_button)

    def __on_forward_primary_press(self, x, y, event):
        """
            Go forward in history
            @param x as int
            @param y as int
            @param even as Gdk.Event
        """
        self.forward()

    def __on_forward_secondary_press(self, x, y):
        """
            Show forward menu
            @param x as int
            @param y as int
        """
        self.__show_menu(self.__forward_button)

    def __on_forward_primary_long(self, x, y):
        """
            Show forward menu
            @param x as int
            @param y as int
        """
        self.__show_menu(self.__forward_button)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from gettext import gettext as _

from eolie.define import App
from eolie.widget_uri_entry import UriEntry


class ToolbarTitle(Gtk.EventBox):
    """
        Title toolbar
    """

    def __init__(self, window):
        """
            Init toolbar
            @param window as Window
        """
        Gtk.EventBox.__init__(self)
        self.__window = window
        self.__width = -1
        self.__input_warning_shown = False
        self.__entry = UriEntry(window)
        self.__entry.show()
        self.add(self.__entry)

    def start_search(self):
        """
            Focus widget without showing
            popover allowing user to start a search
        """
        if not self.__entry.widget.is_focus():
            self.__entry.widget.grab_focus()

    def show_javascript(self, dialog):
        """
            Show a popover with javascript message
            @param dialog as WebKit2.ScriptDialog
        """
        if dialog.get_message():
            from eolie.popover_javascript import JavaScriptPopover
            popover = JavaScriptPopover(dialog, self.__window)
            popover.set_relative_to(self)
            popover.popup()

    def show_geolocation(self, uri, request):
        """
            Show a popover allowing geolocation
            @param uri as str
            @param request as WebKit2.PermissionRequest
        """
        if App().websettings.get("geolocation", uri):
            request.allow()
            self.icons.show_geolocation(True)
        else:
            from eolie.popover_geolocation import GeolocationPopover
            popover = GeolocationPopover(uri, request, self.__window)
            popover.set_relative_to(self)
            popover.popup()

    def show_message(self, message):
        """
            Show a message to user
            @param webview as WebView
            @param msg as str
        """
        from eolie.popover_message import MessagePopover
        popover = MessagePopover(message, self.__window)
        popover.set_relative_to(self)
        popover.popup()

    def show_input_warning(self, webview):
        """
            Show a message to user about password input field over HTTP
            @param webview as WebView
        """
        if self.__input_warning_shown:
            return
        self.__input_warning_shown = True
        self.show_message(_(
            "Heads-up: This page is insecure.\n"
            "If you type your password,\nit will be "
            "visible to cybercriminals!"))

    def do_get_preferred_width(self):
        """
            Fixed preferred width
        """
        if self.__width == -1:
            (min_width, nat_width) = Gtk.Bin.do_get_preferred_width(self)
        else:
            nat_width = self.__width
        return (-1, nat_width)

    def set_width(self, width):
        """
            Set Gtk.Entry width
            @param width as int
        """
        self.__width = width
        self.queue_resize()

    @property
    def entry(self):
        """
            Get entry
            @return UriEntry
        """
        return self.__entry

#######################
# PRIVATE             #
#######################



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import gi
gi.require_version("Soup", "3.0")
from gi.repository import GLib, Soup

from threading import Thread

from eolie.logger import Logger


class TaskHelper:
    """
        Simple helper for running a task in background
    """

    def __init__(self, user_agent=None):
        """
            Init helper
            @param user_agent as str
        """
        self.__signals = {}
        self.__headers = []
        self.__user_agent = user_agent

    def add_header(self, name, value):
        """
            Add header
            @param name as str
            @param value as str
        """
        self.__headers.append((name, value))

    def run(self, command, *args, **kwd):
        """
            run command with params and return to callback
            @param command as function
            @param *args as command arguments
            @param **kwd as { "callback": (function, *args) }
        """
        thread = Thread(target=self.__run,
                        args=(command, kwd, *args))
        thread.daemon = True
        thread.start()

    def load_uri_content(self, uri, cancellable, callback, *args):
        """
            Load uri with libsoup (better performance than Gio)
            @param uri as str
            @param cancellable as Gio.Cancellable
            @param callback as a function
            @callback (uri as str, status as bool, content as bytes, args)
        """
        try:
            session = Soup.Session.new()
            session.set_accept_language_auto(True)
            if self.__user_agent is not None:
                session.set_user_agent(self.__user_agent)
            msg = Soup.Message.new("GET", uri)
            session.send_async(msg,
                               GLib.PRIORITY_DEFAULT,
                               cancellable,
                               self.__on_request_send_async,
                               callback,
                               cancellable,
                               uri,
                               *args)
        except Exception as e:
            Logger.error("HelperTask::load_uri_content(): %s, %s:", e, uri)
            callback(None, False, b"", *args)

#######################
# PRIVATE             #
#######################
    def __run(self, command, kwd, *args):
        """
            Pass command result to callback
            @param command as function
            @param *args as command arguments
            @param kwd as { "callback": (function, *args) }
        """
        try:
            result = command(*args)
            if "callback" in kwd.keys():
                (callback, *callback_args) = kwd["callback"]
                if callback is not None:
                    GLib.idle_add(callback, result, *callback_args)
        except Exception as e:
            Logger.error("TaskHelper::__run(): %s, %s, %s", e, command, kwd)

    def __on_read_bytes_async(self, stream, result, content,
                              cancellable, callback, uri, *args):
        """
            Read data from stream, when finished, pass to callback
            @param stream as Gio.InputStream
            @param result as Gio.AsyncResult
            @param cancellable as Gio.Cancellable
            @param content as bytes
            @param callback as function
            @param uri as str
        """
        try:
            content_result = stream.read_bytes_finish(result)
            content_bytes = content_result.get_data()
            if content_bytes:
                content += content_bytes
                stream.read_bytes_async(4096, GLib.PRIORITY_LOW,
                                        cancellable,
                                        self.__on_read_bytes_async,
                                        content, cancellable, callback,
                                        uri, *args)
            else:
                callback(uri, True, bytes(content), *args)
        except Exception as e:
            Logger.error("TaskHelper::__on_read_bytes_async(): %s", e)
            callback(uri, False, b"", *args)

    def __on_request_send_async(self, source, result, callback,
                                cancellable, uri, *args):
        """
            Get stream and start reading from it
            @param source as Soup.Session
            @param result as Gio.AsyncResult
            @param cancellable as Gio.Cancellable
            @param callback as a function
            @param uri as str
        """
        try:
            stream = source.send_finish(result)
            # We use a bytearray here as seems that bytes += is really slow
            stream.read_bytes_async(4096, GLib.PRIORITY_LOW,
                                    cancellable, self.__on_read_bytes_async,
                                    bytearray(0), cancellable, callback, uri,
                                    *args)
        except Exception as e:
            Logger.warning("TaskHelper::__on_request_send_async(): %s", e)
            callback(uri, False, b"", *args)



# Copyright (c) 2017-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GObject, GLib

from hashlib import md5

from eolie.helper_task import TaskHelper
from eolie.define import EOLIE_CACHE_PATH
from eolie.css_stylesheet import StyleSheet
from eolie.logger import Logger


class StyleSheets(GObject.Object):
    """
        Represent a stylesheets collection
    """

    __gsignals__ = {
        "populated": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "not-cached": (GObject.SignalFlags.RUN_FIRST, None, ())
    }

    def __init__(self):
        """
            Init StyleSheet
        """
        GObject.Object.__init__(self)
        self.__task_helper = TaskHelper()
        self.__cancellable = None
        self.__populated = True
        self.__stylesheets = {}

    def set_cancellable(self, cancellable):
        """
            Set current cancellable
            @param cancellable as Gio.Cancellable
        """
        self.__cancellable = cancellable

    def load_css_uri(self, message, started_time):
        """
            Load CSS URI as user style
            @param started_time as int
            @param message as str
        """
        uri = message.replace("@EOLIE_CSS_URI@", "")
        if not uri:
            self.emit("populated")
            return
        if uri in self.__stylesheets.keys():
            self.__stylesheets[uri].set_started_time(started_time)
            self.__check_populated()
        else:
            self.__populated = False
            stylesheet = self.__load_from_cache(uri)
            if stylesheet is None:
                self.emit("not-cached")
                stylesheet = StyleSheet(uri=uri)
                stylesheet.connect("populated", self.__on_stylesheet_populated)
                self.__stylesheets[uri] = stylesheet
                stylesheet.set_started_time(started_time)
                self.__task_helper.run(stylesheet.populate)
            else:
                self.__stylesheets[uri] = stylesheet
                stylesheet.set_started_time(started_time)
                self.__check_populated()

    def load_css_text(self, message, uri, started_time):
        """
            Load CSS text as user style
            @param started_time as int
            @param uri as str
            @param message as str
        """
        contents = message.replace("@EOLIE_CSS_TEXT@", "")
        if contents:
            css_hash = md5(contents.encode("utf-8")).hexdigest()
            if css_hash in self.__stylesheets.keys():
                self.__stylesheets[css_hash].set_started_time(started_time)
                self.__check_populated()
                return
            self.__populated = False
            stylesheet = StyleSheet(uri=uri, contents=contents)
            self.__stylesheets[css_hash] = stylesheet
            stylesheet.set_started_time(started_time)
            stylesheet.connect("populated", self.__on_stylesheet_populated)
            self.__task_helper.run(stylesheet.populate)

    def get_css_text(self, started_time):
        """
            Get css text for all stylsheets
            @param started_time as str
            @return str
        """
        css = []
        for key in list(self.__stylesheets.keys()):
            stylesheet = self.__stylesheets[key]
            if stylesheet.started_time == started_time:
                css.append(stylesheet.css_text)
            else:
                del self.__stylesheets[key]
        return "".join(css)

    def remove_cache(self):
        """
            Remove cache for current stylesheets
        """
        try:
            for uri in self.__stylesheets.keys():
                encoded = md5(uri.encode("utf-8")).hexdigest()
                filepath = "%s/css/%s.css" % (EOLIE_CACHE_PATH, encoded)
                f = Gio.File.new_for_path(filepath)
                if f.query_exists():
                    f.delete(None)
        except Exception as e:
            Logger.error("StyleSheets::remove_cache(): %s", e)

    def reset(self):
        """
            Reset stylesheet state
        """
        self.__populated = False

    @property
    def populated(self):
        """
            True if populated
            @return bool
        """
        return self.__populated

#######################
# PRIVATE             #
#######################
    def __check_populated(self):
        """
            Check all stylsheets are populated and emit signal if so
        """
        if not self.__populated:
            for uri in self.__stylesheets.keys():
                if not self.__stylesheets[uri].populated:
                    return
            self.__populated = True
            GLib.idle_add(self.emit, "populated")

    def __load_from_cache(self, uri):
        """
            Load CSS from cache
            @param uri as str
            @return StyleSheet
        """
        try:
            encoded = md5(uri.encode("utf-8")).hexdigest()
            filepath = "%s/css/%s.css" % (EOLIE_CACHE_PATH, encoded)
            f = Gio.File.new_for_path(filepath)
            if f.query_exists():
                (status, contents, tags) = f.load_contents(None)
                if status:
                    stylesheet = StyleSheet(uri=uri)
                    stylesheet.set_css_text(contents.decode("utf-8"))
                    return stylesheet
        except Exception as e:
            Logger.error("StyleSheets::__load_from_cache(): %s", e)
        return None

    def __on_stylesheet_populated(self, stylesheet):
        """
            Load stylesheet
            @param stylesheet as StyleSheet
        """
        if stylesheet.uri is not None:
            encoded = md5(stylesheet.uri.encode("utf-8")).hexdigest()
            filepath = "%s/css/%s.css" % (EOLIE_CACHE_PATH, encoded)
            f = Gio.File.new_for_path(filepath)
            fstream = f.replace(None, False,
                                Gio.FileCreateFlags.REPLACE_DESTINATION,
                                None)
            if fstream is not None:
                fstream.write(stylesheet.css_text.encode("utf-8"), None)
                fstream.close()
        self.__check_populated()
